"""
Tools to provide meta-interfaces of python objects.
See also:
    py2misc/py2store/tree_store.py
"""
import inspect
from inspect import signature, Parameter
from typing import Mapping

from i2.util import lazyprop, FrozenDict, get_function_body

inspect_empty = Parameter.empty


class NotFoundType:
    def __bool__(self):
        return False

    def __repr__(self):
        return 'NotFound'


# Note, if the "empty" classes the are used to check if an object is empty are not singletons, could lead to bugs
# TODO: Look into metaprogramming tricks for this.
not_found = NotFoundType()
no_name = type('NoName', (NotFoundType,), {})()
inspect_is_empty = inspect._empty


def is_not_empty(obj):
    if obj is inspect_is_empty or isinstance(obj, NotFoundType):
        return False
    else:
        return True


def _is_property(attr_name, attr_val):
    return not attr_name.startswith('_') and isinstance(attr_val, (property, lazyprop))


def _property_names_of(obj):
    if not isinstance(obj, type):
        obj = type(obj)
    for attr_name in dir(obj):
        attr_val = getattr(obj, attr_name)
        if _is_property(attr_name, attr_val):
            yield attr_name


def name_of_obj(o):
    """
    Tries to find the (or "a") name for an object, even if `__name__` doesn't exist.

    >>> name_of_obj(map)
    'map'
    >>> name_of_obj([1, 2, 3])
    'list'
    >>> name_of_obj(print)
    'print'
    >>> name_of_obj(lambda x: x)
    '<lambda>'
    >>> from functools import partial
    >>> name_of_obj(partial(print, sep=','))
    'print'
    """
    if hasattr(o, '__name__'):
        return o.__name__
    elif hasattr(o, '__class__'):
        name = name_of_obj(o.__class__)
        if name == 'partial':
            if hasattr(o, 'func'):
                return name_of_obj(o.func)
        return name
    else:
        return no_name


class AttrFromKey:
    def __init__(self, d):
        self._d = d

    def __getattr__(self, k):
        return self._d.__getitem__(k)

    __getitem__ = __getattr__


class KeyFromAttr:
    def __init__(self, d):
        self._d = d

    def __getitem__(self, k):
        return getattr(self._d, k)

    __getattr__ = __getitem__


class _ParameterMintLazy:  # Note: Experimental
    _attrs = ['name', 'default', 'annotation']

    #     find = DelegatedAttribute('_params', '__getattribute__')
    #     __getattribute__ = DelegatedAttribute('_params', '__getattribute__')
    def __init__(self, param, _attrs=None):
        self._param = param
        if _attrs is not None:
            # usually used to get full attrs: ['name', 'kind', 'default', 'annotation']
            self._attrs = _attrs

    def items(self):
        for k in self._attrs:
            v = getattr(self._param, k)
            if v is not inspect_empty:
                yield k, v

    def __getitem__(self, k):
        return getattr(self._param, k)

    def __contains__(self, k):
        try:
            _ = self[k]
            return True
        except Exception:
            return False

    def __getattr__(self, k):  # TODO: Use descriptor to make it faster
        return getattr(self._param, k)

    def __getstate__(self):
        return {k: v for k, v in self.items()}


class ParameterMint:
    _attrs = ['name', 'kind', 'default', 'annotation']

    def __init__(self, param, position=None):
        if hasattr(param, '__getitem__'):
            param = AttrFromKey(param)

        for attr in self._attrs:
            try:
                val = getattr(param, attr, inspect_empty)
            except KeyError:
                val = inspect_empty
            setattr(self, attr, val)
        if position is not None:
            self.position = position

    def __getitem__(self, k):
        return getattr(self, k)

    def __contains__(self, k):
        try:
            _ = self[k]
            return True
        except Exception:
            return False

    def items(self):
        for k in self._attrs:
            v = getattr(self, k)
            if v is not inspect_empty:
                yield k, v
        if hasattr(self, 'position'):
            yield 'position', self.position

    def __getstate__(self):
        return {k: v for k, v in self.items()}

    def __repr__(self):
        d = dict()
        for k, v in self.items():
            if k == 'kind':
                v = v.name
            d[k] = v
        return str(d)


dflt_params = FrozenDict({})


class ParametersMint(Mapping):
    """
    Get mint of the parameters of a callable.

    >>> import inspect
    >>> from pprint import pprint
    >>>
    >>> def g(a, b: 'some_type', c=1, d: int = 1) -> float:
    ...     return a * b * c * d
    >>> mint = ParametersMint(inspect.signature(g).parameters)
    >>> # mint is a mapping (like a read-only dict), so...
    >>> list(mint)
    ['a', 'b', 'c', 'd']
    >>>
    >>> for arg_spec in mint.values():
    ...     print(arg_spec)
    {'name': 'a', 'kind': 'POSITIONAL_OR_KEYWORD', 'position': 0}
    {'name': 'b', 'kind': 'POSITIONAL_OR_KEYWORD', 'annotation': 'some_type', 'position': 1}
    {'name': 'c', 'kind': 'POSITIONAL_OR_KEYWORD', 'default': 1, 'position': 2}
    {'name': 'd', 'kind': 'POSITIONAL_OR_KEYWORD', 'default': 1, 'annotation': <class 'int'>, 'position': 3}
    >>> t = list(mint.items())
    >>> t[0]
    ('a', {'name': 'a', 'kind': 'POSITIONAL_OR_KEYWORD', 'position': 0})
    >>> t[1]
    ('b', {'name': 'b', 'kind': 'POSITIONAL_OR_KEYWORD', 'annotation': 'some_type', 'position': 1})
    >>>
    >>> mint = ParametersMint(inspect.signature(g).parameters)
    >>> pprint(dict(mint))
    {'a': {'name': 'a', 'kind': 'POSITIONAL_OR_KEYWORD', 'position': 0},
     'b': {'name': 'b', 'kind': 'POSITIONAL_OR_KEYWORD', 'annotation': 'some_type', 'position': 1},
     'c': {'name': 'c', 'kind': 'POSITIONAL_OR_KEYWORD', 'default': 1, 'position': 2},
     'd': {'name': 'd', 'kind': 'POSITIONAL_OR_KEYWORD', 'default': 1, 'annotation': <class 'int'>, 'position': 3}}

    >>>
    >>> # and now, some cannibalistic fun...
    >>> pprint(dict(ParametersMint(inspect.signature(pprint).parameters)))
    {'compact': {'name': 'compact', 'kind': 'KEYWORD_ONLY', 'default': False, 'position': 5},
     'depth': {'name': 'depth', 'kind': 'POSITIONAL_OR_KEYWORD', 'default': None, 'position': 4},
     'indent': {'name': 'indent', 'kind': 'POSITIONAL_OR_KEYWORD', 'default': 1, 'position': 2},
     'object': {'name': 'object', 'kind': 'POSITIONAL_OR_KEYWORD', 'position': 0},
     'sort_dicts': {'name': 'sort_dicts', 'kind': 'KEYWORD_ONLY', 'default': True, 'position': 6},
     'stream': {'name': 'stream', 'kind': 'POSITIONAL_OR_KEYWORD', 'default': None, 'position': 1},
     'width': {'name': 'width', 'kind': 'POSITIONAL_OR_KEYWORD', 'default': 80, 'position': 3}}
    >>> pprint(dict(ParametersMint(inspect.signature(ParametersMint).parameters)))
    {'args': {'name': 'args', 'kind': 'VAR_POSITIONAL', 'position': 0},
     'kwds': {'name': 'kwds', 'kind': 'VAR_KEYWORD', 'position': 1}}
    >>> pprint(dict(ParametersMint(inspect.signature(ParametersMint.__init__).parameters)))
    {'params': {'name': 'params', 'kind': 'POSITIONAL_OR_KEYWORD', 'default': FrozenDict({}), 'position': 1},
     'self': {'name': 'self', 'kind': 'POSITIONAL_OR_KEYWORD', 'position': 0}}
    >>> pprint(dict(ParametersMint(inspect.signature(ParametersMint.__new__).parameters)))
    {'args': {'name': 'args', 'kind': 'VAR_POSITIONAL', 'position': 1},
     'cls': {'name': 'cls', 'kind': 'POSITIONAL_OR_KEYWORD', 'position': 0},
     'kwds': {'name': 'kwds', 'kind': 'VAR_KEYWORD', 'position': 2}}
    """

    def __init__(self, params=dflt_params):
        self._param_names = list()
        for i, (k, v) in enumerate(params.items()):
            self._param_names.append(k)
            setattr(self, k, ParameterMint(v, position=i))

    def __getitem__(self, k):
        return getattr(self, k)

    def __iter__(self):
        yield from self._param_names

    def items(self):
        for k in self._param_names:
            yield k, getattr(self, k)

    def __len__(self):
        c = 0
        for k in self.__iter__():
            c += 1
        return c

    def __getstate__(self):
        return {k: v.__getstate__() for k, v in self.items()}

    def __setstate__(self, state):
        return type(self)(state)

    def to_dict(self):
        return self.__getstate__()

    def __repr__(self):
        arg_repr = str({k: v.__repr__() for k, v in self.items()})
        return f'{self.__class__.__name__}({arg_repr})'


class Mint(Mapping):
    """
    Get a Mint object of a python object.
    A Mint will provide parameters that provide (meta-)information about the interface of the python object.

    >>> from pprint import pprint
    >>> # Mint of a function
    >>> def f(my_arg: int = 7) -> int:
    ...     return my_arg + 10
    >>> mint = Mint(f)
    >>> mint.obj_name, mint.type_name, mint.module_name, mint.obj_name
    ('f', 'function', 'i2.base', 'f')
    >>> # Mint of a module
    >>> import os as myos
    >>> mint = Mint(myos)
    >>> mint.obj_name, mint.type_name, mint.module_name, mint.obj_name
    ('os', 'module', 'os', 'os')
    >>> assert set(list(mint)) == {'module_name', 'module', 'type_name', 'obj_name'}
    >>> # Mint of a variable
    >>> v = 10
    >>> mint = Mint(v)
    >>> mint.obj_name, mint.type_name, mint.module_name, mint.obj_name
    (NotFound, 'int', NotFound, NotFound)
    >>> assert set(list(mint)) == {'type_name'}  # see that there's only one non-null attr!
    """

    def __init__(self, obj, attrs=None):
        self._obj = obj
        if attrs is None:
            attrs = frozenset(_property_names_of(type(self)))
        else:
            self_type = type(self)
            for attr in attrs:
                if not _is_property(attr, getattr(self_type, attr, None)):
                    raise AttributeError(
                        f"{self.__class__} doesn't have attribute: {attr}"
                    )
        self._attrs = attrs

    def __len__(self):
        return self._length

    def __iter__(self):
        yield from self._non_empty_attrs

    def __getitem__(self, k):
        if k in self._non_empty_attrs:
            return getattr(self, k)
        else:
            raise KeyError('No such key: {k}')

    def items(self):
        for attr in self._non_empty_attrs:
            yield attr, getattr(self, attr)

    @lazyprop
    def _non_empty_attrs(self):
        non_empty_attrs = list()
        for attr in self._attrs:
            attr_val = getattr(self, attr)
            if is_not_empty(attr_val):
                non_empty_attrs.append(attr)
        return non_empty_attrs

    @lazyprop
    def _length(self):
        return len(self._non_empty_attrs)

    @lazyprop
    def obj_name(self):
        return getattr(self._obj, '__name__', not_found)

    @lazyprop
    def type_name(self):
        return getattr(type(self._obj), '__name__', not_found)

    @lazyprop
    def module(self):
        return inspect.getmodule(self._obj) or not_found

    @lazyprop
    def module_name(self):
        return getattr(self.module, '__name__', not_found)


class MintOfCallableMixin:
    @lazyprop
    def _signature(self):
        """Here's some doc"""
        return signature(self._obj)

    @lazyprop
    def parameters(self):
        return ParametersMint(self._signature.parameters)

    @lazyprop
    def return_annotation(self):
        return self._signature.return_annotation or not_found

    @lazyprop
    def doc_string(self):
        return inspect.getdoc(self._obj) or not_found

    @lazyprop
    def comments_preceeding_def(self):
        return inspect.getcomments(self._obj) or not_found

    @lazyprop
    def default_of(self):
        d = dict()
        for k, v in self.parameters.items():
            if 'default' in v:
                d[k] = v['default']
        return d

    @lazyprop
    def annotation_of(self):
        d = dict()
        for k, v in self.parameters.items():
            if 'annotation' in v:
                d[k] = v['annotation']
        return d


class MintOfDocMixin:
    @lazyprop
    def _parsed_doc(self):
        return 'Not yet implemented (correctly)'
        from i2.scrap import parse_mint_doc

        # return parse_mint_doc(self.doc_string)


class MintOfCallable(Mint, MintOfCallableMixin, MintOfDocMixin):
    """
    Get a Mint object of a python object.
    A Mint will provide parameters that provide (meta-)information about the interface of the python object.

    >>> from pprint import pprint
    >>> def f(my_arg: int = 7) -> int:
    ...     return my_arg + 10
    >>> f.__doc__ = 'some documentation'
    >>>
    >>> mint = MintOfCallable(f)
    >>> mint.obj_name
    'f'
    >>> mint.type_name
    'function'
    >>> mint.module_name
    'i2.base'
    >>> mint.parameters.my_arg
    {'name': 'my_arg', 'kind': 'POSITIONAL_OR_KEYWORD', 'default': 7, 'annotation': <class 'int'>, 'position': 0}
    >>> mint.doc_string
    'some documentation'
    >>> mint.return_annotation
    <class 'int'>
    >>> def g(a, b: 'some_string_id_of_a_custom_type', c=1, d: int = 1) -> float:
    ...     return a * b * c * d
    >>> pprint(dict(MintOfCallable(g).parameters))
    {'a': {'name': 'a', 'kind': 'POSITIONAL_OR_KEYWORD', 'position': 0},
     'b': {'name': 'b', 'kind': 'POSITIONAL_OR_KEYWORD', 'annotation': 'some_string_id_of_a_custom_type', 'position': 1},
     'c': {'name': 'c', 'kind': 'POSITIONAL_OR_KEYWORD', 'default': 1, 'position': 2},
     'd': {'name': 'd', 'kind': 'POSITIONAL_OR_KEYWORD', 'default': 1, 'annotation': <class 'int'>, 'position': 3}}
    """

    pass


# class FunctionBuilderMint(Mint):
#     name = lazyprop(Mint.obj_name)
#     module = lazyprop(Mint.module_name)
#     module.__doc__ = 'Name of the module from which this function was imported.'
#     doc = lazyprop(MintOfCallableMixin.doc_string)
#
#     @lazyprop
#     def _signature(self):
#         """ Here's some doc """
#         return signature(self._obj)
#
#     @lazyprop
#     def body(self):
#         """String version of the code representing the body
#         of the function. Defaults to ``'pass'``, which will result
#         in a function which does nothing and returns ``None``."""
#         return get_function_body(self._obj)
#
#     @lazyprop
#     def args(self):
#         return [p['name'] for p in self._signature.parameters]
#
#     @lazyprop
#     def varargs(self):
#         return [p['name'] for p in self._signature.parameters]
#
# """    name (str): Name of the function.
#     doc (str): `Docstring`_ for the function, defaults to empty.
#     module (str): Name of the module from which this function was
#         imported. Defaults to None.
#     body (str): String version of the code representing the body
#         of the function. Defaults to ``'pass'``, which will result
#         in a function which does nothing and returns ``None``.
#     args (list): List of argument names, defaults to empty list,
#         denoting no arguments.
#     varargs (str): Name of the catch-all variable for positional
#         arguments. E.g., "args" if the resultant function is to have
#         ``*args`` in the signature. Defaults to None.
#     varkw (str): Name of the catch-all variable for keyword
#         arguments. E.g., "kwargs" if the resultant function is to have
#         ``**kwargs`` in the signature. Defaults to None.
#     defaults (dict): A mapping of argument names to default values.
#     kwonlyargs (list): Argument names which are only valid as
#         keyword arguments. **Python 3 only.**
#     kwonlydefaults (dict): A mapping, same as normal *defaults*,
#         but only for the *kwonlyargs*. **Python 3 only.**
#     annotations (dict): Mapping of type hints and so
#         forth. **Python 3 only.**
#     filename (str): The filename that will appear in
#         tracebacks. Defaults to "boltons.funcutils.FunctionBuilder".
#     indent (int): Number of spaces with which to indent the
#         function *body*. Values less than 1 will result in an error.
#     dict (dict): Any other attributes which should be added to the
#         functions compiled with this FunctionBuilder."""



"""Test utils"""

import pickle


def is_pickable(obj):
    try:
        _ = pickle.loads(pickle.dumps(obj))
        return True
    except Exception:
        return False


def unpickled_func_still_works(func, *args, **kwargs):
    expected = func(*args, **kwargs)
    unpickled_func = pickle.loads(pickle.dumps(func))
    result = unpickled_func(*args, **kwargs)
    return expected == result



"""Test signatures module


# Notes to the reader

Both in the code and in the docs, we'll use short hands for parameter (argument) kind.
    PK = Parameter.POSITIONAL_OR_KEYWORD
    VP = Parameter.VAR_POSITIONAL
    VK = Parameter.VAR_KEYWORD
    PO = Parameter.POSITIONAL_ONLY
    KO = Parameter.KEYWORD_ONLY

"""
import pytest
from functools import reduce
from typing import Any

from i2 import Sig
from i2.signatures import *
from i2.signatures import (
    normalized_func,
    sigs_for_sigless_builtin_name,
    _robust_signature_of_callable,
)

from i2.tests.util import (
    call_and_return_error,
    sig_to_inputs,
    trace_call,
    function_is_compatible_with_signature,
)


def test_signature_of_partial():
    from functools import partial

    def foo(a, b, c=3) -> int:
        return a + b * c

    assert str(Sig(foo)) == '(a, b, c=3) -> int'
    assert str(Sig(partial(foo, 1))) == '(b, c=3) -> int'
    assert str(Sig(partial(foo, a=1, b=2))) == '(*, a=1, b=2, c=3) -> int'
    assert str(Sig(partial(foo, 1, 2))) == '(c=3) -> int'
    assert str(Sig(partial(foo, 1, b=2))) == '(*, b=2, c=3) -> int'


def test_some_edge_cases_of_sig():
    from operator import itemgetter, attrgetter, methodcaller

    assert Sig(itemgetter).names == ['key', 'keys']
    assert Sig(itemgetter(1)).names == ['iterable']
    assert Sig(itemgetter(1, 2)).names == ['iterable']
    assert Sig(attrgetter).names == ['key', 'keys']
    assert Sig(attrgetter('foo')).names == ['iterable']
    assert Sig(attrgetter('foo', 'bar')).names == ['iterable']
    assert Sig(methodcaller).names == ['name', 'args', 'kwargs']
    # assert Sig(methodcaller('foo')).names == []  # fix!!


def test_sig_wrap_edge_cases():
    """Tests some edge cases involving simultaneous changes of defaults and kinds.

    Current Sig.wrap design allows you to change defaults, and this will have the
    effect of changing the ``__defaults__`` and ``__kwdefaults__`` of the wrapped
    function.

    But, when we change the default of a parameter and change it's kind at the same
    time, bad things happen (see https://github.com/i2mint/i2/issues/16).

    So `Sig.wrap` makes a few checks and raises an error if it's thinks it's not
    safe to do the wrapping.

    Some of these facts may change with new designs. When this is the case,
    this test should be invalidated."""

    # Non edge-case tests:

    def foo(x, y, z=0):
        return x + y * z

    assert foo.__defaults__ == (0,)
    assert foo(1, 2) == 1

    @Sig(lambda x, y, z=3: None)
    def foo(x, y, z=0):
        return x + y * z

    assert Sig(foo)._defaults_ == (3,)

    assert foo(1, 2) == 7
    #    works because Sig also changed __defaults__:
    assert foo.__defaults__ == (3,)

    @Sig(lambda x, y, *, z=3: None)
    def foo(x, y, *, z=0):
        return x + y * z

    assert foo(1, 2) == 7
    #    works because Sig also changed __defaults__ and __kwdefaults__:
    assert foo.__defaults__ == ()
    assert foo.__kwdefaults__ == {'z': 3}
    assert Sig(foo)._defaults_ == ()
    assert Sig(foo)._kwdefaults_ == {'z': 3}

    # The following (where we go from a (same kind) param not having a default,
    # to having one, also work

    @Sig(lambda x, y, z=3: None)
    def foo(x, y, z):
        return x + y * z

    assert foo(1, 2) == 7
    assert foo(1, 2, z=10) == 21

    @Sig(lambda x, y, *, z=3: None)
    def foo(x, y, *, z):
        return x + y * z

    assert foo(1, 2) == 7
    assert foo(1, 2, z=10) == 21
    assert Sig(foo)._defaults_ == ()
    assert Sig(foo)._kwdefaults_ == {'z': 3}


def test_tuple_the_args():
    from i2.signatures import tuple_the_args

    def func(a, *args, bar):
        return trace_call(func, locals())

    assert func(1, 2, 3, bar=4) == 'func(a=1, args=(2, 3), bar=4)'

    wfunc = tuple_the_args(func)

    # here, not that (1) args is specified as one iterable ([2, 3] instead of 2,
    # 3) and (2) the function name is the same as the wrapped (func)
    assert wfunc(1, [2, 3], bar=4) == 'func(a=1, args=(2, 3), bar=4)'

    # See the func itself hasn't changed
    assert func(1, 2, 3, bar=4) == 'func(a=1, args=(2, 3), bar=4)'

    assert str(Sig(func)) == '(a, *args, bar)'
    # See that args is now a PK kind with a default of (). Also, bar became KO.
    assert str(Sig(wfunc)) == '(a, args=(), *, bar)'

    # -----------------------------------------------------------------------------------
    # Let's see what happens when we give bar a default value

    def func2(a, *args, bar=10):
        return trace_call(func2, locals())

    wfunc = tuple_the_args(func2)
    assert wfunc(1, [2, 3]) == 'func2(a=1, args=(2, 3), bar=10)'
    assert wfunc(1, [2, 3], bar=4) == 'func2(a=1, args=(2, 3), bar=4)'

    # On the other hand, specifying bar as a positional won't work.
    # The reason is: args was a variadic, so everything after it should be KO or VK
    # The tuple_the_args doesn't change those signatures.
    #
    with pytest.raises(FuncCallNotMatchingSignature) as e_info:
        wfunc(1, [2, 3], 4)
        assert e_info.value == (
            'There should be only keyword arguments after the Variadic args. '
            'Function was called with (positional=(1, [2, 3], 4), keywords={})'
        )

    # pytest.raises()


@pytest.mark.xfail
def test_normalize_func_simply(function_normalizer=normalized_func):
    # -----------------------------------------------------------------------------------
    def p0113(po1, /, pk1, pk2, *, ko1):
        return f'{po1=}, {pk1=}, {pk2=}, {ko1=}'

    func = p0113
    po1, pk1, pk2, ko1 = 1, 2, 3, 4

    norm_func = function_normalizer(func)

    func_output = func(po1, pk1, pk2=pk2, ko1=ko1)

    norm_func_output = norm_func(po1, pk1, pk2, ko1)

    assert norm_func_output == func_output
    norm_func_output = norm_func(po1, pk1, pk2, ko1=ko1)
    assert norm_func_output == func_output
    norm_func_output = norm_func(po1, pk1, pk2=pk2, ko1=ko1)
    assert norm_func_output == func_output
    norm_func_output = norm_func(po1, pk1=pk1, pk2=pk2, ko1=ko1)
    assert norm_func_output == func_output
    norm_func_output = norm_func(po1=po1, pk1=pk1, pk2=pk2, ko1=ko1)
    assert norm_func_output == func_output

    # -----------------------------------------------------------------------------------
    def p1234(pka, *vpa, koa, **vka):
        return f'{pka=}, {vpa=}, {koa=}, {vka=}'

    pka, vpa, koa, vka = 1, (2, 3), 4, {'a': 'b', 'c': 'd'}

    func = p1234
    norm_func = function_normalizer(func)

    func_output = func(pka, *vpa, koa, **vka)
    norm_func_output = norm_func(pka, vpa, koa, vka)

    assert norm_func_output == func_output


# -----------------------------------------------------------------------------------


def p1234(pka, *vpa, koa, **vka):
    return f'{pka=}, {vpa=}, {koa=}, {vka=}'


@pytest.mark.xfail
def test_normalize_func_combinatorially(function_normalizer=normalized_func):
    # -----------------------------------------------------------------------------------
    def p0113(po1, /, pk1, pk2, *, ko1):
        return f'{po1=}, {pk1=}, {pk2=}, {ko1=}'

    func = p0113
    po1, pk1, pk2, ko1 = 1, 2, 3, 4

    poa = [po1]
    ppka, kpka = [pk1], {'pk2': pk2}
    vpa = []  # no VP argument
    koa = {'ko1': ko1}
    vka = {}  # no VK argument

    norm_func = function_normalizer(func)

    func_output = func(*poa, *ppka, *vpa, **kpka, **koa, **vka)
    norm_func_output = norm_func(po1, pk1, pk2, ko1)
    assert norm_func_output == func_output
    norm_func_output = norm_func(po1, pk1, pk2, ko1=ko1)
    assert norm_func_output == func_output
    norm_func_output = norm_func(po1, pk1, pk2=pk2, ko1=ko1)
    assert norm_func_output == func_output
    norm_func_output = norm_func(po1, pk1=pk1, pk2=pk2, ko1=ko1)
    assert norm_func_output == func_output
    norm_func_output = norm_func(po1=po1, pk1=pk1, pk2=pk2, ko1=ko1)
    assert norm_func_output == func_output

    # -----------------------------------------------------------------------------------
    def p1234(pka, *vpa, koa, **vka):
        return f'{pka=}, {vpa=}, {koa=}, {vka=}'

    pka, vpa, koa, vka = 1, (2, 3), 4, {'a': 'b', 'c': 'd'}

    func = p1234
    norm_func = function_normalizer(func)

    func_output = func(pka, *vpa, koa, **vka)
    norm_func_output = norm_func(pka, vpa, koa, vka)

    assert norm_func_output == func_output

    # -----------------------------------------------------------------------------------


# TODO: It seems in some cases, the better choice would be to oblige the user to deal
#  with return annotation explicitly


def mk_sig(
    obj: Union[Signature, Callable, Mapping, None] = None,
    return_annotations=empty,
    **annotations,
):
    """Convenience function to make a signature or inject annotations to an existing one.

    >>> s = mk_sig(lambda a, b, c=1, d='bar': ..., b=int, d=str)
    >>> s
    <Signature (a, b: int, c=1, d: str = 'bar')>
    >>> # showing that sig can take a signature input, and overwrite an existing annotation:
    >>> mk_sig(s, a=list, b=float)  # note the b=float
    <Signature (a: list, b: float, c=1, d: str = 'bar')>
    >>> mk_sig()
    <Signature ()>

    Trying to annotate an argument that doesn't exist will lead to an AssertionError:

    >>> mk_sig(lambda a, b=2, c=3: ..., d=int)  # doctest: +SKIP
    Traceback (most recent call last):
    ...
    AssertionError: These argument names weren't found in the signature: {'d'}
    """
    if obj is None:
        return Signature()
    if callable(obj):
        obj = Signature.from_callable(obj)  # get a signature object from a callable
    if isinstance(obj, Signature):
        obj = obj.parameters  # get the parameters attribute from a signature
    params = dict(obj)  # get a writable copy of parameters
    if not annotations:
        return Signature(params.values(), return_annotation=return_annotations)
    else:
        assert set(annotations) <= set(
            params
        ), f"These argument names weren't found in the signature: {set(annotations) - set(params)}"
        for name, annotation in annotations.items():
            p = params[name]
            params[name] = Parameter(
                name=name, kind=p.kind, default=p.default, annotation=annotation,
            )
        return Signature(params.values(), return_annotation=return_annotations)


def mk_signature(parameters, *, return_annotation=empty, __validate_parameters__=True):
    """Make an inspect.Signature object with less boilerplate verbosity.
    Args:
        signature: A list of parameter specifications. This could be an inspect.Parameter object or anything that
            the mk_param function can resolve into an inspect.Parameter object.
        return_annotation: Passed on to inspect.Signature.
        __validate_parameters__: Passed on to inspect.Signature.

    Returns:
        An inspect.Signature object

    # >>> mk_signature(['a', 'b', 'c'])
    # <Signature (a, b, c)>
    # >>> mk_signature(['a', ('b', None), ('c', 42, int)])  # specifying defaults and annotations
    # <Signature (a, b=None, c: int = 42)>
    # >>> import inspect
    # >>> mk_signature(['a', ('b', inspect._empty, int)])  # specifying an annotation without a default
    # <Signature (a, b: int)>
    # >>> mk_signature(['a', 'b', 'c'], return_annotation=str)  # specifying return annotation
    # <Signature (a, b, c) -> str>
    # >>>
    # >>> # But you can always specify parameters the "long" way
    # >>> mk_signature([inspect.Parameter(name='kws', kind=inspect.Parameter.VAR_KEYWORD)], return_annotation=str)
    # <Signature (**kws) -> str>
    # >>>
    # >>> # Note that mk_signature is an inverse of signature_to_dict:
    # >>> def foo(a, b: int=0, c=None) -> int: ...
    # >>> sig_foo = signature(foo)
    # >>> assert mk_signature(**signature_to_dict(sig_foo)) == sig_foo

    """
    return Sig(parameters, return_annotation=return_annotation)


# PATTERN: tree crud pattern
def signature_to_dict(sig: Signature):
    # warn("Use Sig instead", DeprecationWarning)
    # return Sig(sig).to_simple_signature()
    return {
        'parameters': sig.parameters,
        'return_annotation': sig.return_annotation,
    }


def _merge_sig_dicts(sig1_dict, sig2_dict):
    """Merge two signature dicts. A in dict.update(sig1_dict, **sig2_dict),
    but specialized for signature dicts.
    If sig1_dict and sig2_dict both define a parameter or return annotation,
    sig2_dict decides on what the output is.
    """
    return {
        'parameters': dict(sig1_dict['parameters'], **sig2_dict['parameters']),
        'return_annotation': sig2_dict['return_annotation']
        or sig1_dict['return_annotation'],
    }


def _merge_signatures(sig1, sig2):
    """Get the merged signatures of two signatures (sig2 is the final decider of conflics)

    >>> def foo(a='a', b: int=0, c=None) -> int: ...
    >>> def bar(b: float=0.0, d: str='hi') -> float: ...
    >>> foo_sig = signature(foo)
    >>> bar_sig = signature(bar)
    >>> foo_sig
    <Signature (a='a', b: int = 0, c=None) -> int>
    >>> bar_sig
    <Signature (b: float = 0.0, d: str = 'hi') -> float>
    >>> _merge_signatures(foo_sig, bar_sig)
    <Signature (a='a', b: float = 0.0, c=None, d: str = 'hi') -> float>
    >>> _merge_signatures(bar_sig, foo_sig)
    <Signature (b: int = 0, d: str = 'hi', a='a', c=None) -> int>
    """
    # sig1_dict = Sig(sig1).to_simple_signature()
    # sig1_dict = signature_to_dict(sig1)
    # # remove variadic kinds from sig1
    # sig1_dict['parameters'] = {k: v for k, v in sig1_dict['parameters'].items() if v.kind not in var_param_kinds}
    # return Sig(**_merge_sig_dicts(sig1_dict, Sig(sig2).to_simple_dict()))
    sig1_dict = signature_to_dict(sig1)
    # remove variadic kinds from sig1
    sig1_dict['parameters'] = {
        k: v
        for k, v in sig1_dict['parameters'].items()
        if v.kind not in var_param_kinds
    }
    kws = _merge_sig_dicts(sig1_dict, signature_to_dict(sig2))
    kws['obj'] = kws.pop('parameters')
    return Sig(**kws).to_simple_signature()


def _merge_signatures_of_funcs(func1, func2):
    """Get the merged signatures of two functions (func2 is the final decider of conflics)

    >>> def foo(a='a', b: int=0, c=None) -> int: ...
    >>> def bar(b: float=0.0, d: str='hi') -> float: ...
    >>> _merge_signatures_of_funcs(foo, bar)
    <Signature (a='a', b: float = 0.0, c=None, d: str = 'hi') -> float>
    >>> _merge_signatures_of_funcs(bar, foo)
    <Signature (b: int = 0, d: str = 'hi', a='a', c=None) -> int>
    """
    return _merge_signatures(signature(func1), signature(func2))


def _merged_signatures_of_func_list(funcs, return_annotation: Any = empty):
    """

    >>> def foo(a='a', b: int=0, c=None) -> int: ...
    >>> def bar(b: float=0.0, d: str='hi') -> float: ...
    >>> def hello(x: str='hi', y=1) -> str: ...
    >>>
    >>> # Test how the order of the functions affect the order of the parameters
    >>> _merged_signatures_of_func_list([foo, bar, hello])
    <Signature (a='a', b: float = 0.0, c=None, d: str = 'hi', x: str = 'hi', y=1)>
    >>> _merged_signatures_of_func_list([hello, foo, bar])
    <Signature (x: str = 'hi', y=1, a='a', b: float = 0.0, c=None, d: str = 'hi')>
    >>> _merged_signatures_of_func_list([foo, bar, hello])
    <Signature (a='a', b: float = 0.0, c=None, d: str = 'hi', x: str = 'hi', y=1)>
    >>>
    >>> # Test the return_annotation argument
    >>> _merged_signatures_of_func_list([foo, bar], list)  # specifying that the return type is a list
    <Signature (a='a', b: float = 0.0, c=None, d: str = 'hi') -> list>
    >>> _merged_signatures_of_func_list([foo, bar], foo)  # specifying that the return type is a list
    <Signature (a='a', b: float = 0.0, c=None, d: str = 'hi') -> int>
    >>> _merged_signatures_of_func_list([foo, bar], bar)  # specifying that the return type is a list
    <Signature (a='a', b: float = 0.0, c=None, d: str = 'hi') -> float>
    """

    s = reduce(_merge_signatures, map(signature, funcs))
    # s = Sig.from_objs(*funcs).to_simple_signature()

    if (
        return_annotation in funcs
    ):  # then you want the return annotation of a specific func of funcs
        return_annotation = signature(return_annotation).return_annotation

    return s.replace(return_annotation=return_annotation)


# TODO: will we need more options for the priority argument? Like position?
def update_signature_with_signatures_from_funcs(*funcs, priority: str = 'last'):
    """Make a decorator that will merge the signatures of given funcs to the signature of the wrapped func.
    By default, the funcs signatures will be placed last, but can be given priority by asking priority = 'first'

    >>> def foo(a='a', b: int=0, c=None) -> int: ...
    >>> def bar(b: float=0.0, d: str='hi') -> float: ...
    >>> def something(y=(1, 2)): ...
    >>> def another(y=10): ...
    >>> @update_signature_with_signatures_from_funcs(foo, bar)
    ... def hello(x: str='hi', y=1) -> str:
    ...     pass
    >>> signature(hello)
    <Signature (x: str = 'hi', y=1, a='a', b: float = 0.0, c=None, d: str = 'hi')>
    >>>
    >>> # Try a different order and priority == 'first'. Notice the b arg type and default!
    >>> add_foobar_to_signature_first = update_signature_with_signatures_from_funcs(
    ...     bar, foo, priority='first'
    ... )
    >>> bar_foo_something = add_foobar_to_signature_first(something)
    >>> signature(bar_foo_something)
    <Signature (b: int = 0, d: str = 'hi', a='a', c=None, y=(1, 2))>
    >>> # See how you can reuse the decorator several times
    >>> bar_foo_another = add_foobar_to_signature_first(another)
    >>> signature(bar_foo_another)
    <Signature (b: int = 0, d: str = 'hi', a='a', c=None, y=10)>
    """
    if not isinstance(priority, str):
        raise TypeError('priority should be a string')

    if priority == 'last':

        def transform_signature(func):
            # func.__signature__ = Sig.from_objs(func, *funcs).to_simple_signature()
            func.__signature__ = _merged_signatures_of_func_list([func] + list(funcs))
            return func

    elif priority == 'first':

        def transform_signature(func):
            # func.__signature__ = Sig.from_objs(*funcs, func).to_simple_signature()
            func.__signature__ = _merged_signatures_of_func_list(list(funcs) + [func])
            return func

    else:
        raise ValueError("priority should be 'last' or 'first'")

    return transform_signature


@pytest.mark.parametrize(
    'sig_spec',
    [
        ('(po, /)'),
        ('(po=0, /)'),
        ('(pk)'),
        ('(pk=0)'),
        ('(*, ko)'),
        ('(*, ko=0)'),
        ('(po, /, pk, *, ko)'),
        ('(po=0, /, pk=0, *, ko=0)'),
        ('(*args)'),
        ('(**kwargs)'),
        ('(*args, **kwargs)'),
        ('(po, /, pk, *args, ko)'),
        ('(po=0, /, pk=0, *args, ko=0)'),
        ('(po, /, pk, *, ko, **kwargs)'),
        ('(po=0, /, pk=0, *, ko=0, **kwargs)'),
        ('(po, /, pk, *args, ko, **kwargs)'),
        ('(po=0, /, pk=0, *args, ko=0, **kwargs)'),
        ('(po1, po2, /)'),
        ('(pk1, pk2)'),
        ('(*, ko1, ko2)'),
        ('(po1, po2, /, pk1, pk2, *, ko1, ko2)'),
        ('(po1, po2, /, pk1, pk2, *args, ko1, ko2, **kwargs)'),
        ('(po1=0, po2=0, /, pk1=0, pk2=0, *args, ko1=0, ko2=0, **kwargs)'),
    ],
)
def test_call_forgivingly(sig_spec):
    sig = Sig(sig_spec)

    @sig
    def foo(*args, **kwargs):
        return args, kwargs

    def validate_call_forgivingly(*args, **kwargs):
        expected_output_kwargs = (
            kwargs
            if VK in sig.kinds.values()
            else {k: v for k, v in kwargs.items() if k in sig}
        )
        pk_in_kwargs_count = sum(
            kind == PK
            for param, kind in sig.kinds.items()
            if param in expected_output_kwargs
        )
        expected_output_args_count = (
            len(args)
            if VP in sig.kinds.values()
            else sum(kind <= PK for kind in sig.kinds.values()) - pk_in_kwargs_count
        )
        expected_output_args = args[:expected_output_args_count]
        expected_output = (expected_output_args, expected_output_kwargs)
        output = call_forgivingly(foo, *args, **kwargs)
        # print()
        # print(args, kwargs)
        # print(expected_output)
        # print(output)
        assert output == expected_output

    for args, kwargs in sig_to_inputs(sig, variadics_source=((), {})):
        kwargs = dict(kwargs, **{'some': 'extra', 'added': 'kwargs'})
        po_pk_count = sum(kind <= PK for kind in sig.kinds.values())
        if len(args) == po_pk_count:
            args = args + ('some', 'extra', 'args')

        validate_call_forgivingly(*args, **kwargs)


@pytest.mark.parametrize(
    'sig_spec1, sig_spec2',
    [
        ('()', '(a)'),
        ('()', '(a=0)'),
        ('(a, /, *, c)', '(a, /, b, *, c)'),
        ('(a, /, *, c)', '(a, /, b=0, *, c)'),
        ('(a, /, b)', '(a, /, b, *, c)'),
        ('(a, /, b)', '(a, /, b, *, c=0)'),
        ('(a, /, b, *, c)', '(*args)'),
        ('(a, /, b, *, c)', '(**kwargs)'),
        ('(a, /, b, *, c)', '(*args, **kwargs)'),
        ('(a, /, b, *, c)', '(a, /, b, *args, c, **kwargs)'),
        ('(a, /, b, *, c)', '(a, b, c)'),
        ('(a, /, b, *, c)', '(a, b, /, *, c)'),
        ('(a, /, b, *, c)', '(a, /, *, b, c)'),
        ('(a, /, b, *, c)', '(a, b, /, c)'),
        ('(a, /, b, *, c)', '(a, *, b, c)'),
        ('(a, /, b, *, c)', '(x, /, b, *, c)',),
        ('(a, /, b, *, c)', '(a, /, x, *, c)',),
        ('(a, /, b, *, c)', '(a, /, b, *, x)',),
        ('(a, /, b, *, c)', '(a=0, b=0, c=0)'),
        ('(a=0, /, b=0, *, c=0)', '(a, b, c)'),
        ('(a, b, /, c, d, *, e, f)', '(b, a, /, d, c, *, f, e)'),
        ('(a, b, /, c, d, *, e, f)', '(a, c, /, b, e, *, d, f)',),
        ('()', '(*args)'),
        ('()', '(**kwargs)'),
        ('()', '(*args, **kwargs)'),
        ('(*args)', '()'),
        ('(*args)', '(**kwargs)'),
        ('(*args)', '(*args, **kwargs)'),
        ('(**kwargs)', '()'),
        ('(**kwargs)', '(*args)'),
        ('(**kwargs)', '(*args, **kwargs)'),
        ('(*args, **kwargs)', '(*args, **kwargs)'),
    ],
)
def test_call_compatibility(sig_spec1, sig_spec2):
    sig1 = Sig(sig_spec1)
    sig2 = Sig(sig_spec2)
    is_compatible = sig1.is_call_compatible_with(sig2)

    exec_env = dict()
    f_def = f'def f{sig_spec2}: pass'
    exec(f_def, exec_env)
    foo = exec_env['f']

    # @sig2
    # def foo(*args, **kwargs):
    #     pass

    pos1, pks1, vp1, kos1, vk1 = sig1.detail_names_by_kind()
    for args, kwargs in sig_to_inputs(sig1, variadics_source=((), {})):
        if vp1 is not None and len(args) == len(pos1) + len(pks1):
            args += ('extra_arg',)
        if vk1 is not None:
            kwargs['extra_kwarg_key'] = 'extra_kwarg_value'
        try:
            foo(*args, **kwargs)
        except TypeError:
            if is_compatible:
                raise
            else:
                return
    if not is_compatible:
        raise AssertionError('sig1 is not compatible with sig2, when it should.')


def test_bool():
    name = 'bool'

    sig = Sig(sigs_for_sigless_builtin_name[name])

    assert function_is_compatible_with_signature(bool, sig)


def visualize_errors_for_function_call(func, sig):
    """
    Calls func on sig_to_inputs and prints error if any
    """

    print(f'============================================================{str(func)}')
    for args, kwargs in sig_to_inputs(sig):
        result = call_and_return_error(func, *args, **kwargs)
        if result is not None:
            print(f'{args=} , {kwargs=}')
            print(result)


def test_sigless_builtins():
    from operator import itemgetter, attrgetter, methodcaller

    for name in sigs_for_sigless_builtin_name:
        # removed breakpoint as it triggers a pdb session
        if name in ['breakpoint']:
            continue
        sig = Sig(sigs_for_sigless_builtin_name[name])
        assert function_is_compatible_with_signature(eval(name), sig)


if __name__ == '__main__':
    from operator import itemgetter, attrgetter, methodcaller

    for name in sigs_for_sigless_builtin_name:
        # removed breakpoint as it triggers a pdb session
        if name in ['breakpoint']:
            continue
        sig = Sig(sigs_for_sigless_builtin_name[name])
        visualize_errors_for_function_call(eval(name), sig)



"""Analyzing what attributes of an input object a function actually uses"""

import ast
import inspect
import dis
from contextlib import contextmanager
from functools import reduce
from importlib import import_module
import os
from collections import namedtuple
from inspect import getsource, getsourcefile

Import = namedtuple('Import', ['module', 'name', 'alias'])


# TODO: Generalize attrs_used_by_method to attrs_used_by_func.


def _get_ast_root_from(o):
    source_str = None
    source_filepath = None
    if isinstance(o, str) and os.path.isfile(o):
        source_filepath = o
        with open(source_filepath) as fh:
            source_str = fh.read()
    elif not isinstance(o, ast.AST):  # not an AST node...
        source_filepath = getsourcefile(o)
        source_str = getsource(o)
        if not isinstance(source_filepath, str) and isinstance(source_str, str):
            raise ValueError('Unrecognized object format')

    return ast.parse(source=source_str, filename=source_filepath)


def _get_imports_from_ast_root(ast_root, recursive=False):
    for node in ast.iter_child_nodes(ast_root):
        module = None
        if isinstance(node, ast.Import):
            module = []
        elif isinstance(node, ast.ImportFrom):
            module = node.module.split('.')

        if module is not None:
            for n in node.names:
                yield Import(module, n.name.split('.'), n.asname)

        if recursive:
            yield from _get_imports_from_ast_root(node, recursive=recursive)


def get_imports_from_obj(o, recursive=False):
    """Getting imports for an object (usually, module)"""
    root = _get_ast_root_from(o)
    yield from _get_imports_from_ast_root(root, recursive)


def _process_duplicates(a, remove_duplicates=True):
    if remove_duplicates:
        return set(a)
        ## remove duplicates conserving order
        # return list(dict.fromkeys(a))
    else:
        return a


def get_class_that_defined_method(method):
    """
    Get class for unbound/bound method.
    """
    if inspect.ismethod(method):
        for cls in inspect.getmro(method.__self__.__class__):
            if cls.__dict__.get(method.__name__) is method:
                return cls
        method = method.__func__  # fallback to __qualname__ parsing
    if inspect.isfunction(method):
        cls = getattr(
            inspect.getmodule(method),
            method.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0],
        )
        if isinstance(cls, type):
            return cls
    return getattr(method, '__objclass__', None)


def cls_and_method_name_of_method(method):
    if isinstance(method, property):
        return get_class_that_defined_method(method.fget), method.fget.__name__
    return get_class_that_defined_method(method), method.__name__


def _alt_cls_and_method_name_of_method(
    method,
):  # TODO: Delete when determined to be of no additional value
    method_path = method.__qualname__.split('.')
    name = method_path[-1]
    cls = reduce(getattr, method_path[:-1], import_module(method.__module__))
    return cls, name


def list_func_calls(fn):
    """
    Extracts functions and methods called from fn
    :param fn: reference to function or method
    :return: a list of functions or methods names
    """
    funcs = []
    bytecode = dis.Bytecode(fn)
    instrs = list(reversed([instr for instr in bytecode]))
    for (ix, instr) in enumerate(instrs):
        if instr.opname == 'CALL_FUNCTION' or instr.opname == 'CALL_METHOD':
            load_func_instr = instrs[ix + instr.arg + 1]
            funcs.append(load_func_instr.argval)
    return [funcname for funcname in reversed(funcs)]


def attr_list(root, func_name):
    """
    Extracts attributes from ast tree processing only func_name function or method
    :param root: root node of ast tree
    :param func_name: name of the function
    :return: a list of attributes names
    """
    atts = []
    functions = sorted(
        {node.name for node in ast.walk(root) if isinstance(node, ast.FunctionDef)}
    )
    for root in ast.walk(root):
        if isinstance(root, ast.FunctionDef) and root.name == func_name:
            for child in ast.walk(root):
                if (
                    isinstance(child, ast.Attribute)
                    and isinstance(child.ctx, ast.Load)
                    and child.attr not in functions
                ):
                    atts.append(child.attr)

    return atts


def _attrs_used_by_method(cls, method_name, remove_duplicates=True):
    """
    Util for attrs_used_by_method. Same output as attrs_used_by_method, but intput is (cls, method_name)
    """
    f = open(inspect.getsourcefile(cls), 'r')
    if f.mode == 'r':
        src = f.read()
    root = ast.parse(src)
    funcs = list_func_calls(getattr(cls, method_name))
    attrs = []
    for func in funcs:
        attrs = attrs + attr_list(root, func)

    return _process_duplicates(
        attrs + attr_list(root, method_name), remove_duplicates=True
    )


def attrs_used_by_method(method, remove_duplicates=True):
    """
    Extracts a list of cls attributes which are used by a method or method_name function
    Args:
        method: The method (object) to analyze

    Returns:
        A list of attribute names (of the class or instance thereof) that are accessed in the code of the said method.

    Example:

    Consider the method `A.target_method` coming from the following code in
    `i2.tests.footprints_test`:
    ```python
    def func(obj):
        \"\"\"Accesses attributes 'a' and 'b' of obj\"\"\"
        return obj.a + obj.b

    class A:
        e = 2

        def __init__(self, a=1, b=0, c=1, d=10):
            self.a = a
            self.b = b
            self.c = c
            self.d = d

        def target_method(self, x):
            \"\"\"Accesses ['a', 'b', 'c', 'e']\"\"\"
            t = func(self)  # and this function will access some attributes!
            tt = self.other_method(t)
            return x * tt / self.e

        def other_method(self, x=1):
            \"\"\"Accesses ['c', 'e']\"\"\"
            w = self.c * 2  # c is accessed first
            return self.e + self.c * x - w  # and c is accessed again

        @classmethod
        def a_class_method(cls, y):
            \"\"\"Accesses ['e']\"\"\"
            return cls.e + y
    ```

    >>> from i2.tests.footprints_test import A
    >>> assert attrs_used_by_method(A.target_method) == {'a', 'b', 'c', 'e'}
    """
    return _attrs_used_by_method(
        *cls_and_method_name_of_method(method), remove_duplicates=remove_duplicates
    )


########## Dyanamic version #################################################################
class Tracker:
    """
    Tracks the attribute access right after `start_track` is set to True.

    Add this to __metaclass__ for any class that you need to track attributes for given a
    target method.
    """

    start_track = False

    def __init__(self, *args, **kwargs):
        self.attr_used = []
        self.attrs_to_ignore = []
        super().__init__(*args, **kwargs)

    def __getattribute__(self, item):
        """
        Inspect getter for tracking the attributes accessed.
        """
        if item not in ['on_access']:
            self.on_access(item)
        return super().__getattribute__(item)

    def __setattr__(self, key, value):
        """
        Inspect setter for tracking the attributes accessed.
        """
        if self.start_track:
            self.on_access(key)

        super().__setattr__(key, value)

    def on_access(self, key):
        """
        on attribute access, record attribute if and only if its not from
        core attribute or `attrs_to_ignore` set to class.
        """
        if (
            key in ['start_track', 'attr_used', 'on_access', 'attrs_to_ignore']
            or key in self.attrs_to_ignore
        ):
            return
        if self.start_track:
            self.attr_used.append(key)


@contextmanager
def start_tracking(tracker_instance):
    """
    Ctx manager to gracefully start/stop tracking.
    """
    tracker_instance.start_track = True
    yield tracker_instance
    tracker_instance.start_track = False


def attrs_used_by_method_computation(
    cls_method, init_kwargs=None, method_kwargs=None, remove_duplicates=True
):
    """
    Tracks the access to attributes within an execution.
    """
    if init_kwargs is None:
        init_kwargs = {}
    if method_kwargs is None:
        method_kwargs = {}

    method_class, method_name = cls_and_method_name_of_method(cls_method)
    tracker = type(
        'Tracker',
        (Tracker, method_class),
        dict(method_class.__dict__, **Tracker.__dict__),
    )(**init_kwargs)
    tracker.attrs_to_ignore = [
        func for func in dir(tracker) if callable(getattr(tracker, func))
    ]

    if hasattr(tracker, method_name):
        # Now, we want to track attributes.
        with start_tracking(tracker):
            if isinstance(cls_method, property):
                candidate_method = cls_method.fget
                candidate_method(tracker)
            else:
                candidate_method = getattr(tracker, method_name)
                candidate_method(**method_kwargs)

        return _process_duplicates(
            tracker.attr_used, remove_duplicates=remove_duplicates
        )

    else:
        # Error class/obj do not have that method.
        return 1


# def test_attrs_used_by_method():
#     def func(obj):
#         return obj.a + obj.b
#
#     class A:
#         e = 2
#
#         def __init__(self, a=0, b=0, c=1, d=10):
#             self.a = a
#             self.b = b
#             self.c = c
#             self.d = d
#
#         def target_func(self, x=3):
#             t = func(self)
#             tt = self.other_method(t)
#             return x * tt / self.e
#
#         target_method = target_func
#
#         def other_method(self, x=1):
#             return self.c * x
#
#     from i2.footprints import attrs_used_by_method
#
#     print(attrs_used_by_method(A.target_func))



"""
Experimental Mapping the maps keys using a switch_case structure.

Development paused while waiting for 3.10's pattern matching that may subsume this.


>>> from i2.switch_case_tree import *
>>> from collections import Counter
>>>
>>> special_featurizer = {
...     'len': len,
...     'cols': lambda df: df.columns,
...     'sum': lambda df: df.sum().sum(),
... }
>>> special_comparison = {
...     'alleq': lambda x, y: all(x == y),
...     'isin': lambda x, y: x in y,
...     'eq': operator.eq,
... }
>>> featurizer = ChainMap(
...     special_featurizer,
...     special_comparison,
...     {'some_local_func': lambda x: list(map(str, x))}
... )
>>>
>>>
>>> comparison = ChainMap(
...     special_comparison, AttrMap(operator, is_valid_val=is_valid_comparision),
... )
>>>
>>> assert comparison['contains'] == operator.contains
>>>

Let's have a look at the "featurizers" (where we purposely injected 3 functions that
were in fact not featurizers!

>>> sorted(featurizer)  # doctest: +NORMALIZE_WHITESPACE
['alleq', 'cols', 'eq', 'isin', 'len', 'some_local_func', 'sum']
>>> Counter(map(is_valid_featurizer, featurizer.values()))
Counter({True: 4, False: 3})
>>> Counter(map(is_valid_comparision, featurizer.values()))
Counter({False: 4, True: 3})

Let's have a look at the comparison functions...

>>> Counter(map(is_valid_comparision, comparison.values()))
Counter({True: 81})
>>> Counter(map(is_valid_featurizer, comparison.values()))
Counter({False: 80, True: 1})

What's that single comparison function that's also a featurizer?

>>> next(filter(is_valid_featurizer, comparison.values())).__name__
'length_hint'


>>> from contextlib import suppress
>>>
>>> with suppress(ModuleNotFoundError, ImportError):
...     import pandas as pd
...     from collections import namedtuple
...     Condition = namedtuple('Condition', ['feat', 'comp'])
...     condition = {
...         feat + '_' + comp: Condition(featurizer[feat], comparison[comp])
...         for feat, comp in [('len', 'lt'), ('cols', 'isin'), ('cols', 'contains'),]
...     }
...     assert all(
...         is_valid_feat_and_comp(feat, comp) for feat, comp in condition.values()
...     )
...
...     df = pd.DataFrame({'a': [1, 2, 3], 'b': [10, 20, 30]})
...     filt = mk_filt(df, *condition['len_lt'])
...     result = list(filter(filt, [2, 3, 4, 5]))
...     assert result == [4, 5]

"""
from collections import ChainMap
from collections.abc import Mapping
import operator
import inspect


class AsIsMap:
    def __init__(self, is_valid_key=None):
        if is_valid_key is None:

            def is_valid_key(x):
                return True

        self._is_valid_key = is_valid_key

    def _validate_key(self, k):
        if not self._is_valid_key(k):
            raise KeyError(f"{k} wasn't a valid key")

    def __getitem__(self, k):
        self._validate_key(k)
        return k


class AttrMap(Mapping):
    def __init__(self, obj, is_valid_val=None):
        self._obj = obj
        if is_valid_val is None:

            def is_valid_val(x):
                return True

        self._is_valid_val = is_valid_val

    @classmethod
    def _validate_key(cls, k):
        if not isinstance(k, str):
            raise KeyError('key should be a string')

    def _validate_val(self, v):
        if not self._is_valid_val(v):
            raise ValueError("key was valid and value found, but value wasn't valid")

    def _getitem(self, k):
        return getattr(self._obj, k)

    def _val_of_key_is_valid(self, k):
        return self._is_valid_val(self._getitem(k))

    def __getitem__(self, k):
        self._validate_key(k)
        v = self._getitem(k)
        self._validate_val(v)
        return v

    def __contains__(self, k):
        self._validate_key(k)
        return hasattr(self._obj, k) and self._val_of_key_is_valid(k)

    def __iter__(self):
        return filter(self._val_of_key_is_valid, dir(self._obj))

    def __len__(self):
        c = 0
        for _ in self.__iter__():
            c += 1
        return c


def n_args_and_n_args_with_no_defaults(func):
    try:
        args = inspect.signature(func).parameters.values()
    except ValueError:  # happens because some builtins don't have signatures (!?!?)
        return 0, 0
    return len(args), len(list(filter(lambda x: x.default == x.empty, args)))


def is_valid_featurizer(func):
    if not callable(func):
        return False
    n_args, n_no_dflt_args = n_args_and_n_args_with_no_defaults(func)
    return (n_args >= 1) and (n_no_dflt_args <= 1)


def is_valid_comparision(func):
    if not callable(func):
        return False
    n_args, n_no_dflt_args = n_args_and_n_args_with_no_defaults(func)
    return (n_args >= 2) and (n_no_dflt_args <= 2)


def is_valid_feat_and_comp(feat, comp):
    return is_valid_featurizer(feat) and is_valid_comparision(comp)


def values_are_valid_feat_and_comp(d):
    return all(is_valid_feat_and_comp(feat, comp) for feat, comp in d.values())


def mk_filt(obj, featurizer, comparison):
    feature = featurizer(obj)

    def filt(x):
        return comparison(feature, x)

    return filt


def test_switch_case_tree():
    from collections import Counter

    special_featurizer = {
        'len': len,
        'cols': lambda df: df.columns,
        'sum': lambda df: df.sum().sum(),
    }

    some_local_func = lambda x: list(map(str, x))

    featurizer = ChainMap(
        special_featurizer,
        {
            k: v
            for k, v in locals().items()
            if is_valid_featurizer(v) and getattr(v, '__module__', '').startswith('i2.')
        },
    )

    special_comparison = {
        'alleq': lambda x, y: all(x == y),
        'isin': lambda x, y: x in y,
        'eq': operator.eq,
    }

    comparison = ChainMap(
        special_comparison, AttrMap(operator, is_valid_val=is_valid_comparision),
    )

    assert comparison['contains'] == operator.contains

    assert Counter(map(is_valid_featurizer, featurizer.values())) == Counter({True: 3})

    assert Counter(map(is_valid_comparision, featurizer.values())) == Counter(
        {False: 3}
    )
    assert Counter(map(is_valid_featurizer, comparison.values())) == Counter(
        {False: 80, True: 1}
    )
    assert Counter(map(is_valid_comparision, comparison.values())) == Counter(
        {True: 81}
    )

    featurizer_kvs = {k: v.__name__ for k, v in featurizer.items()}
    assert featurizer_kvs == {'len': 'len', 'cols': '<lambda>', 'sum': '<lambda>'}

    from contextlib import suppress

    with suppress(ModuleNotFoundError, ImportError):
        import pandas as pd
        from collections import namedtuple

        Condition = namedtuple('Condition', ['feat', 'comp'])
        condition = {
            feat + '_' + comp: Condition(featurizer[feat], comparison[comp])
            for feat, comp in [('len', 'lt'), ('cols', 'isin'), ('cols', 'contains'),]
        }
        assert all(
            is_valid_feat_and_comp(feat, comp) for feat, comp in condition.values()
        )

        df = pd.DataFrame({'a': [1, 2, 3], 'b': [10, 20, 30]})
        filt = mk_filt(df, *condition['len_lt'])
        assert list(filter(filt, [2, 3, 4, 5])) == [4, 5]


##########################################################################################################
# from anytree import Node, AnyNode, RenderTree, ContStyle, NodeMixin
#
# NodeMixin.separator = '.'
#
#
# def get_attr_and_vals(obj):
#     for k in dir(obj):
#         v = getattr(obj, k)
#         yield k, v
#
#
# def attr_and_vals_dict(obj, filt=None):
#     return {k: v for k, v in filter(filt, get_attr_and_vals(obj))}
#
#
# # Filters
# def value_not_callable(kv):
#     return not callable(kv[1])
#
#
# class AttrTreeImporter:
#     _base_types = (int, float, str, bytes, bool, list)
#
#     def __init__(self, val_types=()):
#         self.val_types = tuple(set(val_types).union(self._base_types))
#
#     def dict_of_obj(self, obj):
#         d = dict()
#         for k, v in obj.__dict__.items():
#             if isinstance(v, self.val_types):
#                 d[k] = v
#             else:
#                 d[k] = self.dict_of_obj(v)
#         return d
#
#     def tree_of_dict(self, d, parent=None, root_name='root'):
#         if parent is None:
#             parent = Node(root_name)
#         for k, v in d.items():
#             if isinstance(v, dict):
#                 n = Node(k, parent=parent)
#                 self.tree_of_dict(v, parent=n)
#             else:
#                 n = Node(k, parent=parent, val=v)
#         return parent
#
#     def tree_of_obj(self, obj, root_name=None):
#         if root_name is None:
#             if hasattr(obj, '__name__'):
#                 root_name = obj.__name__
#             elif hasattr(obj, '__class__'):
#                 root_name = obj.__class__.__name__
#             else:
#                 root_name = 'root'
#         return self.tree_of_dict(self.dict_of_obj(obj), root_name=root_name)



"""Meta-programming tools to build declarative frameworks"""
from i2.deco import (
    preprocess,
    postprocess,
    preprocess_arguments,
    input_output_decorator,
    wrap_class_methods_input_and_output,
    double_up_as_factory,
)

from i2.signatures import (
    Sig,
    Param,
    sort_params,
    call_forgivingly,
    call_somewhat_forgivingly,
    name_of_obj,
    empty as empty_param_attr,
)

from i2.multi_object import (
    MultiObj,
    MultiFunc,
    Pipe,
    FuncFanout,
    FlexFuncFanout,
    ParallelFuncs,
    ContextFanout,
)

from i2.errors import (
    InterruptWithBlock,
    HandleExceptions,
)

from i2.wrapper import (
    wrap,
    Wrap,
    Wrapx,
    Ingress,
    include_exclude,
    rm_params,
    partialx,
    ch_names,
    func_to_method_func,
    bind_funcs_object_attrs,
    kwargs_trans,
)

from i2.util import (
    Literal,
    path_extractor,
    get_function_body,
    lazyprop,
    frozendict,
    inject_method,
    mk_sentinel,
    ensure_identifiers,
)



"""
A module containing a bunch of objects for testing.

See also: https://github.com/i2mint/py2http/blob/master/py2http/tests/objects_for_testing.py
"""


###############################################################################
# Empty callables, meant to give an example of different kinds of callables ###
# Note that x can be applied to a number, string, or list #####################
def f(x):
    """Note that x can be applied to a number, string, or list"""
    return x * 2


class C:
    @classmethod
    def c(cls, x):
        return f(x)

    @staticmethod
    def s(x):
        return f(x)

    # instance method
    def i(self, x):
        return f(x)

    # making an instance itself callable
    def __call__(self, x):
        return f(x)


###############################################################################


def times_2(x):
    return x * 2


def plus_1(z):
    return z + 1


def add(a, b: float = 0.0) -> float:
    """Adds numbers"""
    return a + b


def mult(x: float, y=1):
    return x * y


def formula1(w, /, x: float, y=1, *, z: int = 1):
    """Has 3 kinds & every of the 4 combinations of (default y/n, annotated y/n)"""
    return ((w + x) * y) ** z


def formula2(w, /, x: float, y=2, *, z: int = 3):
    """Has 3 kinds & every of the 4 combinations of (default y/n, annotated y/n)

    >>> formula2(0, 1, 2)
    '(w:=0) + (x:=1) * (y:=2) ** (z:=3) == 8'
    """
    return f'(w:={w}) + (x:={x}) * (y:={y}) ** (z:={z}) == {w + x * y ** z}'


# Only *args and **kwargs. Also has a doc with a doc test
def sum_of_args(*args, **kwargs):
    """Sums the args, and returns a copy of kwargs with the sum of args added to each value.

    >>> sum_of_args(1,2,3, a=0, b=10)
    {'a': 6, 'b': 16}
    """
    t = sum(args)
    return {k: t + v for k, v in kwargs.items()}


def test_func_1(
    any_var,
    a_list: list,
    a_dict: dict,
    an_int: int,
    a_float: float = 3.14,
    a_bool=True,
    a_str: str = None,
) -> str:
    """
    This is the first line,
    continued on the next.

    This function is meant to demo several combinations of type declaration (in signature or/and in docs),
    as well as with or without defaults.

    :param any_var: no type declared in either signature of doc
    :param list a_list: type declared in signature AND doc
    :param a_dict: type declared in signature but NOT in doc
    :param int an_int: type declared in doc but NOT in signature
    :param float a_float: type declared in signature AND doc, and has a default of the desired type
    :param a_bool: type declared in signature but NOT in doc, and has a default of the desired type
    :param str a_str: type declared in signature AND doc, but has a default that is not of that type
    line description
    :return str: Will just return the 'test_func_1 returned' string
    :tags foo, bar

    >>> # here are sometests
    >>> test_func_1('anything', [1, 2], {'this': 0, 'that': 'foo'}, 1, 1.1, a_bool=True, a_str='hello')
    'test_func_1 returned'
    """
    # This is the first comment
    return 'test_func_1 returned'  # this is a comment on the same line as some code


class SomeClass(object):
    pass


class SomeOtherClass(object):
    pass


def test_func_2(
    any_var,
    an_int: int,
    a_float=0.1,
    a_tuple: tuple = (),
    an_obj: SomeClass = SomeClass(),
    another_obj=SomeOtherClass(),
):
    """
    This is the first line,
    continued on the next.

    And this is some more information about the function:
    Blah
    Blah
    Bloo

    :param any_var: foo
    :param an_int: you again?!
    :param a_float:
    :param a_tuple: For real
    :param an_obj: Does nothing
    :param another_obj: Multi-
    line
    :return: Just pi
    :tags any, old, tag
    :keyword one_keyword
    :keyword keyword1, keyword2
    """

    """
    This is the first line,
    continued on the next.

    And this is some more information about the function:
    Blah
    Blah
    Bloo

    :param any_var: this can be of any type
    :param an_int: an int
    :param a_float: a float
    :param a_bool: a bool
    :param a_str: a str
    :param a_list:
    :param a_dict: An example of a continued
    line description
    :return:

    >>> # here are sometests
    >>> test_func_1('anything', 1, 1.1, True, 'hello', [1, 2], {'this': 0, 'that': 'foo'})
    'returned'
    """
    return 3.14


class AClass(object):
    def __init__(self, a=1, b=0):
        """The doc of the __init__"""
        self.a = a
        self.b = b

    def __call__(self, x: float) -> float:
        """The actual callable"""
        return self.a * x + self.b


class TestObj:
    constarg: str = ''

    def __init__(self, constarg: str):
        print(constarg)
        self.constarg = constarg

    def methodnum(self, methodarg1: int) -> int:
        print(methodarg1)
        return methodarg1 + 1

    def methodstr(self, methodarg2: str = 'hi') -> str:
        print(methodarg2)
        return methodarg2 + ' test ' + self.constarg


from contextlib import suppress

with suppress(ModuleNotFoundError, ImportError):
    from numpy import ndarray, matrix
    from pandas import DataFrame, Series

    def a_func_with_df_in_and_sr_out(df: DataFrame) -> Series:
        return df.sum()

    def a_func_with_np_and_df(
        df: DataFrame, sr: Series, arr: ndarray, mat: matrix
    ) -> matrix:
        return (df.sum().sum() + sr.sum() + arr.sum()) * mat



"""Meta-interfaces """

import doctest
from typing import Callable
import re
from itertools import groupby
from inspect import getdoc

MAX_LINE_LENGTH = 72  # https://en.wikipedia.org/wiki/Characters_per_line


def _prefix_lines(s: str, prefix: str = '# ', even_if_empty: bool = False) -> str:
    r"""
    Prefix every line of s with given prefix.

    :param s: String whose lines you want to prefix.
    :param prefix: Desired prefix string. Default is '# ', to have the effect of "commenting out" line
    :param even_if_empty: Whether to prefix empty strings or not.
    :return: A string whose lines have been prefixed.

    >>> _prefix_lines('something to comment out')
    '# something to comment out'
    >>> _prefix_lines('A line you want to prefix', prefix='PREFIX: ')
    'PREFIX: A line you want to prefix'
    >>> print(_prefix_lines('What happens\nif the thing to comment out\nhas multiple lines?'))
    # What happens
    # if the thing to comment out
    # has multiple lines?
    >>> _prefix_lines('')  # see that an empty string is returned as is
    ''
    >>> _prefix_lines('', even_if_empty=True)  # unless you ask for it
    '# '
    """
    if not even_if_empty:
        if len(s) == 0:
            return s
    return '\n'.join(map(lambda x: prefix + x, s.split('\n')))


class ExampleX(doctest.Example):
    """doctest.Example eXtended to have more convenient methods"""

    source_prefix = '>>> '
    source_continuation = '... '

    def __init__(
        self, source, want=None, exc_msg=None, lineno=0, indent=0, options=None,
    ):
        # if source is already a doctest.Example instance, use its properties as args
        if isinstance(source, doctest.Example):
            o = source
            source, want, exc_msg, lineno, indent, options = (
                o.source,
                o.want,
                o.exc_msg,
                o.lineno,
                o.indent,
                o.options,
            )
        super().__init__(source, want, exc_msg, lineno, indent, options)

    @property
    def indent_str(self):
        return self.indent * ' '

    @property
    def _source_str(self):
        indented_continuation = f'{self.indent_str}{self.source_continuation}'
        t = self.source.replace('\n', f'\n{indented_continuation}')
        # remove the trailing '...' if present
        if t.endswith(indented_continuation):
            t = t[: -len(indented_continuation)]
        return t

    def __str__(self):
        want = self.want
        if self.want:
            want = self.indent_str + want
        return self.indent_str + self.source_prefix + self._source_str + want

    def __repr__(self):
        return str(self)


from typing import Sequence


class DoctestBlock(list):
    """A list that (should) contain doctest Example instances"""

    def __init__(self, seq=()):
        super().__init__(map(ExampleX, seq))

    def __str__(self):
        return ''.join(map(str, self))

    def __repr__(self):
        return f'<DoctestBlock with {len(self)} examples>\n' + str(self)


parse_doctest = doctest.DocTestParser().parse


# TODO: Newer, using doctest parser. See what can be merged
#  with non_doctest_lines etc.
def split_text_and_doctests(doc_string: str):
    r"""
    Generates alternating blocks of "text" (string) and "doctest blocks"
    (``DoctestBlock`` instances, which are essentially a list of ``ExampleX`` instances).

    >>> example = '''
    ...     This is to test the doctest splitter.
    ...     Until now, we're in a text block.
    ...     The following is a doctest block:
    ...
    ...     >>> 2 + 3
    ...     5
    ...     >>> t = 5
    ...     >>> tt = 10
    ...
    ...     This is another text block, followed with another doctest block:
    ...
    ...     >>> def foo():
    ...     ...     return 42
    ...     >>> foo()
    ...     42
    ...
    ... '''
    >>>
    >>> blocks = list(split_text_and_doctests(example))

    There are 5 blocks:

    >>> len(blocks)
    5

    The first block is a string, corresponding to explanatory text of the doc string:

    >>> isinstance(blocks[0], str)
    True
    >>> print(blocks[0])
    <BLANKLINE>
    This is to test the doctest splitter.
    Until now, we're in a text block.
    The following is a doctest block:
    <BLANKLINE>
    <BLANKLINE>

    The next block is a ``DoctestBlock`` instance.

    >>> block = blocks[1]
    >>> isinstance(block, DoctestBlock)
    True

    This block has 3 elements (``ExampleX`` instances)

    >>> len(block)
    3

    If you ask for the string representation of this block, you'll get a doctest string:

    >>> str(block)
    '    >>> 2 + 3\n    5\n    >>> t = 5\n    >>> tt = 10\n'

    """

    def is_string(item):
        if isinstance(item, str):
            if item == '':
                return False
            else:
                return True
        else:
            return False

    t = parse_doctest(doc_string)
    g = groupby(t, key=is_string)
    for _is_string, group_data in g:
        if _is_string:
            yield '\n'.join(list(group_data))
        else:
            yield DoctestBlock(filter(None, group_data))


comment_strip_p = re.compile(r'(?m)^ *#.*\n?')
doctest_line_p = re.compile('\s*>>>')
empty_line = re.compile('\s*$')


def non_doctest_lines(doc):
    r"""Generator of lines of the doc string that are not in a doctest scope.

    >>> def _test_func():
    ...     '''Line 1
    ...     Another
    ...     >>> doctest_1
    ...     >>> doctest_2
    ...     line_after_a_doc_test
    ...     another_line_that_is_in_the_doc_test scope
    ...
    ...     But now we're out of a doctest's scope
    ...
    ...     >>> Oh no, another doctest!
    ...     '''
    >>> from inspect import getdoc
    >>>
    >>> list(non_doctest_lines(getdoc(_test_func)))
    ['Line 1', 'Another', "But now we're out of a doctest's scope", '']

    :param doc:
    :return:
    """
    last_line_was_a_doc_test = False
    for line in doc.splitlines():
        if not doctest_line_p.match(line):
            if not last_line_was_a_doc_test:
                yield line
                last_line_was_a_doc_test = False
            else:
                if empty_line.match(line):
                    last_line_was_a_doc_test = False
        else:
            last_line_was_a_doc_test = True


def strip_comments(code):
    code = str(code)
    return comment_strip_p.sub('', code)


def mk_example_wants_callback(source_want_func: Callable[[str, str], Callable]):
    def example_wants_callback(example, *args, **kwargs):
        want = example.want.strip()
        if want:
            source = example.source.strip()
            return source_want_func(source, want, *args, **kwargs)
        else:
            return example.source

    return example_wants_callback


def split_line_comments(s):
    t = s.split('#')
    if len(t) == 1:
        comment = ''
    else:
        s, comment = t
    return s, comment


def _assert_wants(source, want, wrap_func_name=None):
    is_a_multiline = len(source.split('\n')) > 1

    if not is_a_multiline:
        source, comment = split_line_comments(source)
        if wrap_func_name is None:
            t = f'({source}) == {want} #{comment}'
        else:
            t = f'{wrap_func_name}({source}) == {wrap_func_name}({want}) #{comment}'
        if "'" in t and not '"' in t:
            strchr = '"'
            return 'assert {t}, {strchr}{t}{strchr}'.format(t=t, strchr=strchr)
        elif '"' in t and not "'" in t:
            strchr = "'"
            return 'assert {t}, {strchr}{t}{strchr}'.format(t=t, strchr=strchr)
        else:
            return 'assert {t}'.format(t=t)
    else:  # if you didn't return before
        if wrap_func_name is None:
            return f'actual = {source}\nexpected = {want}\nassert actual == expected'
        else:
            return (
                f'actual = {wrap_func_name}({source})\nexpected = {wrap_func_name}({want})\n'
                'assert actual == expected'
            )


def _output_prefix(source, want, prefix='# OUTPUT: '):
    return source + '\n' + prefix + want + '\n'


output_prefix = mk_example_wants_callback(_output_prefix)
assert_wants = mk_example_wants_callback(_assert_wants)

# def example_to_doctest_string(source, want):
#     want.replace()
#     return source +


def doctest_string_trans_lines(
    doctest_obj: doctest.DocTest, example_callback=assert_wants
):
    for example in doctest_obj.examples:
        yield example_callback(example)


def _doctest_string_gen(obj, example_callback, recurse=True):
    doctest_finder = doctest.DocTestFinder(verbose=False, recurse=recurse)
    doctest_objs = doctest_finder.find(obj)
    for doctest_obj in doctest_objs:
        yield from doctest_string_trans_lines(doctest_obj, example_callback)


def doctest_string(obj, example_callback=assert_wants, recurse=True):
    """
    Extract the doctests found in given object.

    :param obj: Object (module, class, function, etc.) you want to extract doctests from.
    :params output_prefix:
    :param recurse: Whether the process should find doctests in the attributes of the object, recursively.
    :return: A string containing the doctests, with output lines prefixed by '# Output:'
    """
    return '\n'.join(_doctest_string_gen(obj, example_callback, recurse=recurse))


from functools import partial

doctest_string.for_output_prefix = partial(
    doctest_string, example_callback=output_prefix
)
doctest_string.for_assert_wants = partial(doctest_string, example_callback=assert_wants)


def doctest_string_print(obj, example_callback=assert_wants, recurse=True):
    """
    Extract the doctests found in given object.
    :param obj: Object (module, class, function, etc.) you want to extract doctests from.
    :param recurse: Whether the process should find doctests in the attributes of the object, recursively.
    :return: A string containing the doctests, with output lines prefixed by '# Output:'
    """
    return print(doctest_string(obj, example_callback, recurse=recurse))


def old_doctest_string(
    obj, output_prefix='# OUTPUT: ', include_attr_without_doctests=False, recurse=True,
):
    """
    Extract the doctests found in given object.
    :param obj: Object (module, class, function, etc.) you want to extract doctests from.
    :param output_prefix:
    :param recurse: Whether the process should find doctests in the attributes of the object, recursively.
    :return: A string containing the doctests, with output lines prefixed by '# Output:'
    """
    doctest_finder = doctest.DocTestFinder(verbose=False, recurse=recurse)
    r = doctest_finder.find(obj)
    s = ''
    for rr in r:
        header = f'# {rr.name} '
        header += '#' * max(0, MAX_LINE_LENGTH - len(header)) + '\n'
        ss = ''
        for example in rr.examples:
            want = example.want
            want = want.strip()
            ss += '\n' + example.source + _prefix_lines(want, prefix=output_prefix)
        if include_attr_without_doctests:
            s += header + ss
        elif len(ss) > 0:  # only append this attr if ss is non-empty
            s += header + ss
    return s


# import sphinx

if __name__ == '__main__':
    print(doctest_string(_prefix_lines))
# # _prefix_lines ########################################################
#
# _prefix_lines('something to comment out')
# # OUTPUT: '# something to comment out'
# _prefix_lines('A line you want to prefix', prefix='PREFIX: ')
# # OUTPUT: 'PREFIX: A line you want to prefix'
# print(_prefix_lines('What happens\nif the thing to comment out\nhas multiple lines?'))
# # OUTPUT: # What happens
# # OUTPUT: # if the thing to comment out
# # OUTPUT: # has multiple lines?
# _prefix_lines('')  # see that an empty string is returned as is
# # OUTPUT: ''
# _prefix_lines('', even_if_empty=True)  # unless you ask for it
# # OUTPUT: '# '



"""Misc util objects"""
from operator import attrgetter, itemgetter
import inspect
import re
import itertools
import sys
import functools
import types
from typing import (
    Mapping,
    Callable,
    Any,
    MutableMapping,
    Union,
    Optional,
    Iterable,
    TypeVar,
)

T = TypeVar('T')


class OverwritesForbidden(ValueError):
    """Raise when a user is not allowed to overwrite a mapping's key"""


def is_lambda(func):
    return getattr(func, '__name__', None) == '<lambda>'


# TODO: Fragile. Make more robust.
def lambda_code(lambda_func):
    func_str = str(inspect.getsourcelines(lambda_func)[0])
    return func_str.strip("['\\n']").split(' = ')[1]


# TODO: Only works with lambdas so either assert function is a lambda on construction
#  or make it work with functions more generally.
class PicklableLambda:
    """
    Wraps a lambda function to make it picklable (and provide it with a name, optionally.

    >>> f = lambda x, y=0: x + y
    >>> ff = PicklableLambda(f)
    >>> import pickle
    >>> fff = pickle.loads(pickle.dumps(ff))
    >>> assert fff(2, 3) == ff(2, 3) == f(2, 3)

    See https://stackoverflow.com/questions/73980648/how-to-transform-a-lambda-function-into-a-pickle-able-function

    """

    def __init__(self, func, name=None):
        self.func = func
        self.__signature__ = inspect.signature(self.func)
        self.__name__ = name or getattr(func, '__name__', type(self).__name__)

    def __getstate__(self):
        return lambda_code(self.func), self.__name__

    def __setstate__(self, state):
        func_code, name = state
        func = eval(func_code)  # scary?
        self.__init__(func, name)

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def __repr__(self):
        return f'<{type(self).__name__}({self.__name__})>'


def ensure_identifiers(
    *objs: Iterable[T],
    get_identfiers: Callable[[T], Iterable[str]] = str.split,
    is_identifier: Callable[[str], bool] = str.isidentifier,
):
    """Ensure an iterable of identifiers

    >>> list(ensure_identifiers('these', 'are', 'valid', 'identifiers'))
    ['these', 'are', 'valid', 'identifiers']

    By default, ``ensure_identifiers`` will apply ``str.split`` to each ``obj`` of
    ``objs`` (assumed to be strings!) so that it can extract identifiers from
    space-separated strings of identifiers:

    >>> list(ensure_identifiers('these are valid identifiers'))
    ['these', 'are', 'valid', 'identifiers']

    You can control this functionality through the ``get_identfiers`` argument, for
    example, disallowing such splitting, or enabling the extraction of identifiers
    from other objects than strings.

    >>> list(ensure_identifiers(
    ...     {'this': 0, 'works': 1}, {'too': 2},
    ...     get_identfiers=list
    ... ))
    ['this', 'works', 'too']

    You can also control the ``is_identifier`` validatation function:

    >>> def less_than_6_chars(s): return len(s) < 6
    >>> list(ensure_identifiers('okay', 'too_long', is_identifier=less_than_6_chars))
    Traceback (most recent call last):
      ...
    ValueError: too_long isn't an identifier according toless_than_6_chars

    """
    for obj in objs:
        for identifier in get_identfiers(obj):
            if is_identifier(identifier):
                yield identifier
            else:
                raise ValueError(
                    f"{identifier} isn't an identifier according to"
                    f"{getattr(is_identifier, '__name__', str(is_identifier))}"
                )


def insert_name_based_objects_in_scope(
    *names,
    factory: Callable[[str], Any],
    scope: MutableMapping,
    allow_overwrites: bool = False,
):
    """
    Make several string-parametrized objects and insert them in a scope (e.g. locals()).

    This is useful when to avoid (error-prone) situations where we want the name we
    assign an object to, to be aligned with it's internal name, such as::

        foo = Factory('foo', ...)
        bar = Factory('bar', ...)
        baz = Factory('baz', ...)

    :param names: Identifier (valid python variable name) strings.
        These are used both as arguments of the ``factory`` and as keys for the
        ``scope`` the object the factory makes will be inserted under.
    :param factory: A function that takes a (valid python identifier) string and
        returns an object parametrized by that string.
    :param scope: The ``MutableMapping`` we want to insert the objects in.
    :param allow_overwrites: Whether the objects we create can overwrite existing
        objects the ``scope`` may already have. If we don't allow overwrites and we
        try to write under an existing key, a ``OverwritesForbidden`` error will be
        raised. This also includes the situation where we have some duplicates in
        ``names``.
    :return: None (this function has the side effect of inserting items in ``scope``.

    One of the (controversal) uses of ``insert_name_based_objects_in_scope`` is to be
    able to make several string-parametrized

    >>> from collections import namedtuple
    >>> from functools import partial
    >>>
    >>> factory = partial(namedtuple, field_names='apple banana')
    >>> insert_namedtuples_in_locals = partial(insert_name_based_objects_in_scope,
    ...     factory=factory, scope=locals(), allow_overwrites=True
    ... )
    >>> insert_namedtuples_in_locals('foo bar', 'baz')

    And now ``foo` exists!

    >>> 'foo' in locals()
    True
    >>> foo(1,2)
    foo(apple=1, banana=2)

    And so does ``bar`` and ``baz``:

    >>> bar(3, banana=4)
    bar(apple=3, banana=4)
    >>> baz(apple=3, banana=4)
    baz(apple=3, banana=4)
    """
    for name in ensure_identifiers(*names):
        if name not in scope or allow_overwrites:
            scope[name] = factory(name)
        else:
            raise OverwritesForbidden(
                f'This key already exisited and is not allowed to be overwritten'
            )


class Literal:
    """An object to indicate that the value should be considered literally.

    >>> t = Literal(42)
    >>> t.get_val()
    42
    >>> t()
    42

    """

    def __init__(self, val):
        self.val = val

    def get_val(self):
        """Get the value wrapped by Literal instance.

        One might want to use ``literal.get_val()`` instead ``literal()`` to get the
        value a ``Literal`` is wrapping because ``.get_val`` is more explicit.

        That said, with a bit of hesitation, we allow the ``literal()`` form as well
        since it is useful in situations where we need to use a callback function to
        get a value.
        """
        return self.val

    __call__ = get_val


def dflt_idx_preprocessor(obj, idx):
    if isinstance(idx, str) and str.isdigit(idx):
        idx = int(idx)
    if isinstance(idx, int) or isinstance(obj, Mapping):
        return obj[idx]
    elif hasattr(obj, idx):
        return getattr(obj, idx)
    else:
        raise KeyError(f"Couldn't extract a {idx} from object {obj}")


def path_extractor(tree, path, getter=dflt_idx_preprocessor, *, path_sep='.'):
    """Get items from a tree-structured object from a sequence of tree-traversal indices.

    :param tree: The object you want to extract values from:
        Can be any object you want, as long as the indices listed by path and how to get
        the items indexed are well specified by ``path`` and ``getter``.
    :param path: An iterable of indices that define how to traverse the tree to get
        to desired item(s). If this iterable is a string, the ``path_sep`` argument
        will be used to transform it into a tuple of string indices.
    :param getter: A ``(tree, idx)`` function that specifies how to extract item ``idx``
        from the ``tree`` object.
    :param path_sep: The string separator to use if ``path`` is a string
    :return: The ``tree`` item(s) referenced by ``path``


    >>> tree = {'a': {'b': [0, {'c': [1, 2, 3]}]}}
    >>> path_extractor(tree, path=['a'])
    {'b': [0, {'c': [1, 2, 3]}]}
    >>> path_extractor(tree, path=['a', 'b'])
    [0, {'c': [1, 2, 3]}]
    >>> path_extractor(tree, path=['a', 'b', 1])
    {'c': [1, 2, 3]}
    >>> path_extractor(tree, path=['a', 'b', 1, 'c'])
    [1, 2, 3]
    >>> path_extractor(tree, path=('a', 'b', 1, 'c', 2))
    3

    You could do the same by specifying the path as a dot-separated string.

    >>> path_extractor(tree, 'a.b.1.c.2')
    3

    You can use any separation you want.

    >>> path_extractor(tree, 'a/b/1/c/2', path_sep='/')
    3

    You can also use `*` to indicate that you want to keep all the nodes of a given
    level.

    >>> tree = {'a': [{'b': [1, 10]}, {'b': [2, 20]}, {'b': [3, 30]}]}
    >>> path_extractor(tree, 'a.*.b.1')
    [10, 20, 30]

    A generalization of `*` is to specify a callable which will be intepreted as
    a filter function.

    >>> tree = {'a': [{'b': 1}, {'c': 2}, {'b': 3}, {'b': 4}]}
    >>> path_extractor(tree, ['a', lambda x: 'b' in x])
    [{'b': 1}, {'b': 3}, {'b': 4}]
    >>> path_extractor(tree, ['a', lambda x: 'b' in x, 'b'])
    [1, 3, 4]
    """
    if isinstance(path, str):
        path = path.split(path_sep)
    if len(path) == 0:
        return tree
    else:
        idx, *path = path  # extract path[0] as idx & update path to path[1:]
        if isinstance(idx, str) and idx == '*':
            idx = lambda x: True  # use a filter function (but filter everything in)
        if callable(idx) and not isinstance(idx, Literal):
            # If idx is a non-literal callable, consider it as a filter to be applied
            # to iter(tree)
            # TODO: https://github.com/i2mint/i2/issues/27
            return [
                path_extractor(sub_tree, path, getter) for sub_tree in filter(idx, tree)
            ]
        else:
            if isinstance(idx, Literal):
                # Use of Literal is meant get out of trouble if we want to use a
                # callable as an actual index, not as a filter.
                idx = idx.get_val()
            tree = getter(tree, idx)
            return path_extractor(tree, path, getter)


# Note: Specialization of path_extractor for Mappings
def dp_get(d, dot_path):
    """Get stuff from a dict (or any Mapping), using dot_paths (i.e. 'foo.bar' instead of
     ['foo']['bar'])

    >>> d = {'foo': {'bar': 2, 'alice': 'bob'}, 3: {'pi': 3.14}}
    >>> assert dp_get(d, 'foo') == {'bar': 2, 'alice': 'bob'}
    >>> assert dp_get(d, 'foo.bar') == 2
    >>> assert dp_get(d, 'foo.alice') == 'bob'
    """
    return path_extractor(d, dot_path, lambda d, k: d[k])


class lazyprop:
    """
    A descriptor implementation of lazyprop (cached property) from David Beazley's "Python Cookbook" book.
    It's

    >>> class Test:
    ...     def __init__(self, a):
    ...         self.a = a
    ...     @lazyprop
    ...     def len(self):
    ...         print('generating "len"')
    ...         return len(self.a)
    >>> t = Test([0, 1, 2, 3, 4])
    >>> t.__dict__
    {'a': [0, 1, 2, 3, 4]}
    >>> t.len
    generating "len"
    5
    >>> t.__dict__
    {'a': [0, 1, 2, 3, 4], 'len': 5}
    >>> t.len
    5
    >>> # But careful when using lazyprop that no one will change the value of a without deleting the property first
    >>> t.a = [0, 1, 2]  # if we change a...
    >>> t.len  # ... we still get the old cached value of len
    5
    >>> del t.len  # if we delete the len prop
    >>> t.len  # ... then len being recomputed again
    generating "len"
    3
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, instance, cls):
        if instance is None:
            return self
        else:
            value = self.func(instance)
            setattr(instance, self.func.__name__, value)
            return value


class FrozenHashError(TypeError):
    pass


class FrozenDict(dict):
    """An immutable dict subtype that is hashable and can itself be used
    as a :class:`dict` key or :class:`set` entry. What
    :class:`frozenset` is to :class:`set`, FrozenDict is to
    :class:`dict`.

    There was once an attempt to introduce such a type to the standard
    library, but it was rejected: `PEP 416 <https://www.python.org/dev/peps/pep-0416/>`_.

    Because FrozenDict is a :class:`dict` subtype, it automatically
    works everywhere a dict would, including JSON serialization.

    """

    __slots__ = ('_hash',)

    def updated(self, *a, **kw):
        """Make a copy and add items from a dictionary or iterable (and/or
        keyword arguments), overwriting values under an existing
        key. See :meth:`dict.update` for more details.
        """
        data = dict(self)
        data.update(*a, **kw)
        return type(self)(data)

    @classmethod
    def fromkeys(cls, keys, value=None):
        # one of the lesser known and used/useful dict methods
        return cls(dict.fromkeys(keys, value))

    def __repr__(self):
        cn = self.__class__.__name__
        return '%s(%s)' % (cn, dict.__repr__(self))

    def __reduce_ex__(self, protocol):
        return type(self), (dict(self),)

    def __hash__(self):
        try:
            ret = self._hash
        except AttributeError:
            try:
                ret = self._hash = hash(frozenset(self.items()))
            except Exception as e:
                ret = self._hash = FrozenHashError(e)

        if ret.__class__ is FrozenHashError:
            raise ret

        return ret

    def __copy__(self):
        return self  # immutable types don't copy, see tuple's behavior

    # block everything else
    def _raise_frozen_typeerror(self, *a, **kw):
        'raises a TypeError, because FrozenDicts are immutable'
        raise TypeError('%s object is immutable' % self.__class__.__name__)

    __setitem__ = __delitem__ = update = _raise_frozen_typeerror
    setdefault = pop = popitem = clear = _raise_frozen_typeerror

    del _raise_frozen_typeerror


frozendict = FrozenDict  # alias to align with frozenset

########################################################################################################################


function_type = type(
    lambda x: x
)  # using this instead of callable() because classes are callable, for instance


class NoDefault(object):
    def __repr__(self):
        return 'no_default'


no_default = NoDefault()


class imdict(dict):
    def __hash__(self):
        return id(self)

    def _immutable(self, *args, **kws):
        raise TypeError('object is immutable')

    __setitem__ = _immutable
    __delitem__ = _immutable
    clear = _immutable
    update = _immutable
    setdefault = _immutable
    pop = _immutable
    popitem = _immutable


def inject_method(self, method_function, method_name=None):
    """
    method_function could be:
        * a function
        * a {method_name: function, ...} dict (for multiple injections)
        * a list of functions or (function, method_name) pairs
    """
    if isinstance(method_function, function_type):
        if method_name is None:
            method_name = method_function.__name__
        setattr(self, method_name, types.MethodType(method_function, self))
    else:
        if isinstance(method_function, dict):
            method_function = [
                (func, func_name) for func_name, func in method_function.items()
            ]
        for method in method_function:
            if isinstance(method, tuple) and len(method) == 2:
                self = inject_method(self, method[0], method[1])
            else:
                self = inject_method(self, method)

    return self


########################################################################################################################


def get_function_body(func):
    source_lines = inspect.getsourcelines(func)[0]
    source_lines = itertools.dropwhile(lambda x: x.startswith('@'), source_lines)
    line = next(source_lines).strip()
    if not line.startswith('def ') and not line.startswith('class'):
        return line.rsplit(':')[-1].strip()
    elif not line.endswith(':'):
        for line in source_lines:
            line = line.strip()
            if line.endswith(':'):
                break
    # Handle functions that are not one-liners
    first_line = next(source_lines)
    # Find the indentation of the first line
    indentation = len(first_line) - len(first_line.lstrip())
    return ''.join(
        [first_line[indentation:]] + [line[indentation:] for line in source_lines]
    )


class ExistingArgument(ValueError):
    pass


class MissingArgument(ValueError):
    pass


def _default_sentinel_repr_method(self):
    return '%s(%r)' % (self.__class__.__name__, self.__name__)


def mk_sentinel(
    name,
    boolean_value: bool = False,
    repr_: Union[str, Callable] = _default_sentinel_repr_method,
    *,
    module: Optional[str] = None,
):
    """Creates and returns a new **instance** of a new class, suitable for usage as a
    "sentinel" since it is a kind of singleton (there can be only one instance of it.)

    A frequent use case for sentinels are where we want to indicate that something is
    missing. Often, we use ``None`` for this, but sometimes ``None`` is a valid value in
    our context (see for example the ``inspect.Parameter.empty`` sentinel to indicate
    that an argument doesn't have a default or annotation).
    Other times, we may want to distinguish different kinds of "nothing".

    ``mk_sentinel`` can help you create such sentinels, takes care of annoying details
    like pickability and allows you to control how to resolve your sentinel to a boolean.

    :param name: The name of your sentinel. Will be used for ``__name__`` attribute.
    :param boolean_value: The boolean value that the sentinel instance should resolve to.
    :param repr_: The method or string that should be used for the repr.
    :param module:
    :return: A sentinel instance

    >>> Empty = mk_sentinel('Empty')
    >>> Empty
    Sentinel('Empty')

    By default, the boolean resolution of a sentinel is ``False``. Meaning:

    >>> Nothing = mk_sentinel('Nothing')
    >>> bool(Nothing)
    False

    This is consistent with ``None``, so that you can check that an object ``x`` is not
    ``Nothing`` by doing ``if x: ...`` or idioms like:

    >>> x = Nothing
    >>> x = x or 'default'
    >>> x
    'default'

    (Though note that in situations where other elements that cast to ``False`` are
    valid values for ``x`` (like ``0``, ``None``, or ``False`` itself), it's safer to use
    ``if x is not Nothing: ...``.)

    Anyway, I digress.
    Point is that in some situations, the semantics  or usage of your sentinel is better
    align with True. You can control what the boolean resolution of your
    sentinel should be through the ``boolean_value`` argument:

    >>> Empty = mk_sentinel('Empty', boolean_value=True)
    >>> bool(Empty)
    True

    You can also control what you see in the repr, specifying a string value;

    >>> Empty = mk_sentinel('undefined', repr_='undefined')
    >>> Empty
    undefined

    or a method;

    >>> Empty = mk_sentinel('Empty', repr_=lambda self: f"<{self.__name__}>")
    >>> Empty
    <Empty>

    And yes, even though we used a lambda here, it's still picklable:

    >>> import pickle

    >>> Empty = mk_sentinel('Empty', repr_='Empty', module=__name__)
    >>> pickle.loads(pickle.dumps(Empty))  # doctest: +SKIP
    Empty

    Talking about pickle, here's some more info on that:

    >>> unpickled_Empty = pickle.loads(pickle.dumps(Empty))  # doctest: +SKIP
    >>> # The unpickled version is "equal" to the original:
    >>> unpickled_Empty == Empty  # doctest: +SKIP
    True
    >>> # the types are the same too:
    >>> type(unpickled_Empty) == type(Empty)    # doctest: +SKIP
    True
    >>>
    >>>

    Note that though two sentinels might have the same name, they're not equal:

    >>> Empty = mk_sentinel('Empty')
    >>> AnotherEmptyWithSameName = mk_sentinel('Empty')
    >>> Empty
    Sentinel('Empty')
    >>> AnotherEmptyWithSameName
    Sentinel('Empty')
    >>> # but...
    >>> AnotherEmptyWithSameName == Empty
    False
    >>> # Note even the types are the same!
    >>> type(AnotherEmptyWithSameName) == type(Empty)
    False

    One thing that makes the pickle work is that we took care of sticking in a
    ``__module__`` for you. ``mk_sentinel`` figures this out by some dark magic
    involving looking into the system's "frames" etc. This may not always work since
    some systems (e.g. ``pypy``) may use different "under-the-hood" methods.

    But if you want to control the value of ``__module__`` yourself, you can, simply
    but indicating what the module of the sentinel is.
    Usually, you'll just specify it as ``module=__name__``, which will stick the
    name of the module you're defining the sentinel in for you!

    >>> MySentinel = mk_sentinel('MySentinel', module=__name__)

    Thanks: Inspired greately from the ``make_sentinel`` function of ``boltons``:
    See https://boltons.readthedocs.io/.

    """

    class Sentinel(object):
        def __init__(self):
            self.__name__ = name

        if callable(repr_):
            __repr__ = repr_
        else:

            def __repr__(self):
                return repr_

        def __reduce__(self):
            return self.__name__

        def __bool__(self):
            return boolean_value

    if module is None:
        # TODO: Try to use something else than hidden _getframe
        # TODO: extract this module resolver so can be reused (_getframe(2)?)
        frame = sys._getframe(1)
        module = frame.f_globals.get('__name__')

    if not module or module not in sys.modules:
        raise ValueError(
            'Pickleable sentinel objects can only be made from top-level module scopes'
        )
    Sentinel.__module__ = module

    return Sentinel()


def _indent(text, margin, newline='\n', key=bool):
    'based on boltons.strutils.indent'
    indented_lines = [
        (margin + line if key(line) else line) for line in text.splitlines()
    ]
    return newline.join(indented_lines)


NO_DEFAULT = mk_sentinel('NO_DEFAULT', boolean_value=False)


from inspect import formatannotation


def inspect_formatargspec(
    args,
    varargs=None,
    varkw=None,
    defaults=None,
    kwonlyargs=(),
    kwonlydefaults={},
    annotations={},
    formatarg=str,
    formatvarargs=lambda name: '*' + name,
    formatvarkw=lambda name: '**' + name,
    formatvalue=lambda value: '=' + repr(value),
    formatreturns=lambda text: ' -> ' + text,
    formatannotation=formatannotation,
):
    """Copy formatargspec from python 3.7 standard library.
    Python 3 has deprecated formatargspec and requested that Signature
    be used instead, however this requires a full reimplementation
    of formatargspec() in terms of creating Parameter objects and such.
    Instead of introducing all the object-creation overhead and having
    to reinvent from scratch, just copy their compatibility routine.
    """

    def formatargandannotation(arg):
        result = formatarg(arg)
        if arg in annotations:
            result += ': ' + formatannotation(annotations[arg])
        return result

    specs = []
    if defaults:
        firstdefault = len(args) - len(defaults)
    for i, arg in enumerate(args):
        spec = formatargandannotation(arg)
        if defaults and i >= firstdefault:
            spec = spec + formatvalue(defaults[i - firstdefault])
        specs.append(spec)
    if varargs is not None:
        specs.append(formatvarargs(formatargandannotation(varargs)))
    else:
        if kwonlyargs:
            specs.append('*')
    if kwonlyargs:
        for kwonlyarg in kwonlyargs:
            spec = formatargandannotation(kwonlyarg)
            if kwonlydefaults and kwonlyarg in kwonlydefaults:
                spec += formatvalue(kwonlydefaults[kwonlyarg])
            specs.append(spec)
    if varkw is not None:
        specs.append(formatvarkw(formatargandannotation(varkw)))
    result = '(' + ', '.join(specs) + ')'
    if 'return' in annotations:
        result += formatreturns(formatannotation(annotations['return']))
    return result


class FunctionBuilder(object):
    """The FunctionBuilder type provides an interface for programmatically
    creating new functions, either based on existing functions or from
    scratch.

    Note: Based on https://boltons.readthedocs.io

    Values are passed in at construction or set as attributes on the
    instance. For creating a new function based of an existing one,
    see the :meth:`~FunctionBuilder.from_func` classmethod. At any
    point, :meth:`~FunctionBuilder.get_func` can be called to get a
    newly compiled function, based on the values configured.

    >>> fb = FunctionBuilder('return_five', doc='returns the integer 5',
    ...                      body='return 5')
    >>> f = fb.get_func()
    >>> f()
    5
    >>> fb.varkw = 'kw'
    >>> f_kw = fb.get_func()
    >>> f_kw(ignored_arg='ignored_val')
    5

    Note that function signatures themselves changed quite a bit in
    Python 3, so several arguments are only applicable to
    FunctionBuilder in Python 3. Except for *name*, all arguments to
    the constructor are keyword arguments.

    Args:
        name (str): Name of the function.
        doc (str): `Docstring`_ for the function, defaults to empty.
        module (str): Name of the module from which this function was
            imported. Defaults to None.
        body (str): String version of the code representing the body
            of the function. Defaults to ``'pass'``, which will result
            in a function which does nothing and returns ``None``.
        args (list): List of argument names, defaults to empty list,
            denoting no arguments.
        varargs (str): Name of the catch-all variable for positional
            arguments. E.g., "args" if the resultant function is to have
            ``*args`` in the signature. Defaults to None.
        varkw (str): Name of the catch-all variable for keyword
            arguments. E.g., "kwargs" if the resultant function is to have
            ``**kwargs`` in the signature. Defaults to None.
        defaults (tuple): A tuple containing default argument values for
            those arguments that have defaults.
        kwonlyargs (list): Argument names which are only valid as
            keyword arguments. **Python 3 only.**
        kwonlydefaults (dict): A mapping, same as normal *defaults*,
            but only for the *kwonlyargs*. **Python 3 only.**
        annotations (dict): Mapping of type hints and so
            forth. **Python 3 only.**
        filename (str): The filename that will appear in
            tracebacks. Defaults to "boltons.funcutils.FunctionBuilder".
        indent (int): Number of spaces with which to indent the
            function *body*. Values less than 1 will result in an error.
        dict (dict): Any other attributes which should be added to the
            functions compiled with this FunctionBuilder.

    All of these arguments are also made available as attributes which
    can be mutated as necessary.

    .. _Docstring: https://en.wikipedia.org/wiki/Docstring#Python

    """

    _argspec_defaults = {
        'args': list,
        'varargs': lambda: None,
        'varkw': lambda: None,
        'defaults': lambda: None,
        'kwonlyargs': list,
        'kwonlydefaults': dict,
        'annotations': dict,
    }

    @classmethod
    def _argspec_to_dict(cls, f):
        argspec = inspect.getfullargspec(f)
        return dict((attr, getattr(argspec, attr)) for attr in cls._argspec_defaults)

    _defaults = {
        'doc': str,
        'dict': dict,
        'is_async': lambda: False,
        'module': lambda: None,
        'body': lambda: 'pass',
        'indent': lambda: 4,
        'annotations': dict,
        'filename': lambda: 'boltons.funcutils.FunctionBuilder',
    }

    _defaults.update(_argspec_defaults)

    _compile_count = itertools.count()

    def __init__(self, name, **kw):
        self.name = name
        for a, default_factory in self._defaults.items():
            val = kw.pop(a, None)
            if val is None:
                val = default_factory()
            setattr(self, a, val)

        if kw:
            raise TypeError('unexpected kwargs: %r' % kw.keys())
        return

    # def get_argspec(self):  # TODO

    def get_sig_str(self, with_annotations=True):
        """Return function signature as a string.

        with_annotations is ignored on Python 2.  On Python 3 signature
        will omit annotations if it is set to False.
        """
        if with_annotations:
            annotations = self.annotations
        else:
            annotations = {}

        return inspect_formatargspec(
            self.args, self.varargs, self.varkw, [], self.kwonlyargs, {}, annotations
        )

    _KWONLY_MARKER = re.compile(
        r'''
    \*     # a star
    \s*    # followed by any amount of whitespace
    ,      # followed by a comma
    \s*    # followed by any amount of whitespace
    ''',
        re.VERBOSE,
    )

    def get_invocation_str(self):
        kwonly_pairs = None
        formatters = {}
        if self.kwonlyargs:
            kwonly_pairs = dict((arg, arg) for arg in self.kwonlyargs)
            formatters['formatvalue'] = lambda value: '=' + value

        sig = inspect_formatargspec(
            self.args,
            self.varargs,
            self.varkw,
            [],
            kwonly_pairs,
            kwonly_pairs,
            {},
            **formatters,
        )
        sig = self._KWONLY_MARKER.sub('', sig)
        return sig[1:-1]

    @classmethod
    def from_func(cls, func):
        """Create a new FunctionBuilder instance based on an existing
        function. The original function will not be stored or
        modified.
        """
        # TODO: copy_body? gonna need a good signature regex.
        # TODO: might worry about __closure__?
        if not callable(func):
            raise TypeError('expected callable object, not %r' % (func,))

        if isinstance(func, functools.partial):
            kwargs = {
                'name': func.__name__,
                'doc': func.__doc__,
                'module': getattr(func, '__module__', None),  # e.g., method_descriptor
                'annotations': getattr(func, '__annotations__', {}),
                'dict': getattr(func, '__dict__', {}),
            }

        kwargs.update(cls._argspec_to_dict(func))

        if inspect.iscoroutinefunction(func):
            kwargs['is_async'] = True

        return cls(**kwargs)

    def get_func(self, execdict=None, add_source=True, with_dict=True):
        """Compile and return a new function based on the current values of
        the FunctionBuilder.

        Args:
            execdict (dict): The dictionary representing the scope in
                which the compilation should take place. Defaults to an empty
                dict.
            add_source (bool): Whether to add the source used to a
                special ``__source__`` attribute on the resulting
                function. Defaults to True.
            with_dict (bool): Add any custom attributes, if
                applicable. Defaults to True.

        To see an example of usage, see the implementation of
        :func:`~boltons.funcutils.wraps`.
        """
        execdict = execdict or {}
        body = self.body or self._default_body

        tmpl = 'def {name}{sig_str}:'
        tmpl += '\n{body}'

        if self.is_async:
            tmpl = 'async ' + tmpl

        body = _indent(self.body, ' ' * self.indent)

        name = self.name.replace('<', '_').replace('>', '_')  # lambdas
        src = tmpl.format(
            name=name,
            sig_str=self.get_sig_str(with_annotations=False),
            doc=self.doc,
            body=body,
        )
        self._compile(src, execdict)
        func = execdict[name]

        func.__name__ = self.name
        func.__doc__ = self.doc
        func.__defaults__ = self.defaults
        func.__kwdefaults__ = self.kwonlydefaults
        func.__annotations__ = self.annotations

        if with_dict:
            func.__dict__.update(self.dict)
        func.__module__ = self.module
        # TODO: caller module fallback?

        if add_source:
            func.__source__ = src

        return func

    def get_defaults_dict(self):
        """Get a dictionary of function arguments with defaults and the
        respective values.
        """
        ret = dict(
            reversed(list(zip(reversed(self.args), reversed(self.defaults or []))))
        )
        kwonlydefaults = getattr(self, 'kwonlydefaults', None)
        if kwonlydefaults:
            ret.update(kwonlydefaults)
        return ret

    def get_arg_names(self, only_required=False):
        arg_names = tuple(self.args) + tuple(getattr(self, 'kwonlyargs', ()))
        if only_required:
            defaults_dict = self.get_defaults_dict()
            arg_names = tuple([an for an in arg_names if an not in defaults_dict])
        return arg_names

    def add_arg(self, arg_name, default=NO_DEFAULT, kwonly=False):
        """Add an argument with optional *default* (defaults to
        ``funcutils.NO_DEFAULT``). Pass *kwonly=True* to add a
        keyword-only argument
        """
        if arg_name in self.args:
            raise ExistingArgument(
                'arg %r already in func %s arg list' % (arg_name, self.name)
            )
        if arg_name in self.kwonlyargs:
            raise ExistingArgument(
                'arg %r already in func %s kwonly arg list' % (arg_name, self.name)
            )
        if not kwonly:
            self.args.append(arg_name)
            if default is not NO_DEFAULT:
                self.defaults = (self.defaults or ()) + (default,)
        else:
            self.kwonlyargs.append(arg_name)
            if default is not NO_DEFAULT:
                self.kwonlydefaults[arg_name] = default
        return

    def remove_arg(self, arg_name):
        """Remove an argument from this FunctionBuilder's argument list. The
        resulting function will have one less argument per call to
        this function.

        Args:
            arg_name (str): The name of the argument to remove.

        Raises a :exc:`ValueError` if the argument is not present.

        """
        args = self.args
        d_dict = self.get_defaults_dict()
        try:
            args.remove(arg_name)
        except ValueError:
            try:
                self.kwonlyargs.remove(arg_name)
            except (AttributeError, ValueError):
                # py2, or py3 and missing from both
                exc = MissingArgument(
                    'arg %r not found in %s argument list:'
                    ' %r' % (arg_name, self.name, args)
                )
                exc.arg_name = arg_name
                raise exc
            else:
                self.kwonlydefaults.pop(arg_name, None)
        else:
            d_dict.pop(arg_name, None)
            self.defaults = tuple([d_dict[a] for a in args if a in d_dict])
        return

    def _compile(self, src, execdict):

        filename = '<%s-%d>' % (self.filename, next(self._compile_count),)
        try:
            code = compile(src, filename, 'single')
            exec(code, execdict)
        except Exception:
            raise
        return execdict



"""Flattening maps and manipulating key paths"""

from functools import reduce, wraps, partial
import operator
from collections.abc import MutableMapping


def trans_generator_output(trans):
    def decorator(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            yield from (trans(x) for x in func(*args, **kwargs))

        return wrapped

    return decorator


class NoDefault:
    pass


NO_DFLT = NoDefault()


def flatten_dict(d, sep=None, prefix=''):
    """
    Computes a "flat" dict from a nested one. A flat dict's keys are the paths of the input dict.
    These paths will be expressed as tuples of the original keys by defaults.
    If these keys are strings though, you can use sep and prefix to get string representations of the paths.

    :param d: a nested dict
    :param sep: The separator character (or string) in a string representation of the paths.
    :param prefix: A string to prepend on all the paths
    :return: A flat dict

    >>> d = {'a': {
    ...         'a': '2a',
    ...         'c': {'a': 'aca', 'u': 4}
    ...         },
    ...      'c': 3
    ...     }
    >>> flatten_dict(d)
    {('a', 'a'): '2a', ('a', 'c', 'a'): 'aca', ('a', 'c', 'u'): 4, ('c',): 3}
    >>> flatten_dict(d, sep='.')
    {'a.a': '2a', 'a.c.a': 'aca', 'a.c.u': 4, 'c': 3}
    >>> flatten_dict(d, sep='/', prefix='/ROOT/')
    {'/ROOT/a/a': '2a', '/ROOT/a/c/a': 'aca', '/ROOT/a/c/u': 4, '/ROOT/c': 3}
    """
    if sep is None:
        kp = KeyPathMap(d)
    else:
        kp = StrKeyPath(d, sep=sep, prefix=prefix)
    return kp.flat_dict()


def rollout_dict(d, sep=None, prefix=''):
    """
    Get the nested path of a flat (key path) dict. This is the inverse of flatten_dict.

    :param d: a flat dict (i.e. one whose keys are paths of a nested dict)
    :param sep: If None (default), the paths should be key tuples. If a string, it it assumed to be
        the separator of string representations of the path
    :param prefix: A string that has be prepended to all each key (path) of the input dict
        (and therefore should be removed)
    :return: The corresponding nested path

    >>> flat_d = {('a', 'a'): '2a', ('a', 'c', 'a'): 'aca', ('a', 'c', 'u'): 4, ('c',): 3}
    >>> rollout_dict(flat_d)
    {'a': {'a': '2a', 'c': {'a': 'aca', 'u': 4}}, 'c': 3}
    >>> flat_d = {'a.a': '2a', 'a.c.a': 'aca', 'a.c.u': 4, 'c': 3}
    >>> rollout_dict(flat_d, sep='.')
    {'a': {'a': '2a', 'c': {'a': 'aca', 'u': 4}}, 'c': 3}
    >>> flat_d = {'/ROOT/a/a': '2a', '/ROOT/a/c/a': 'aca', '/ROOT/a/c/u': 4, '/ROOT/c': 3}
    >>> rollout_dict(flat_d, sep='/', prefix='/ROOT/')
    {'a': {'a': '2a', 'c': {'a': 'aca', 'u': 4}}, 'c': 3}
    """
    if sep is None:
        kp = KeyPathMap()
    else:
        kp = StrKeyPath(sep=sep, prefix=prefix)
    return kp.rollout(d)


# Note: Might need to extend control of setitem: overwrites, leaf->node, node->leaf
# Note: Could also control deletes more.
class KeyPathMap(MutableMapping):
    """
    Provides a key-path view to a nested mapping (by default, a dict).
    A nested mapping can be see as a tree, where if a value is itself a mapping, it is a non-terminal node,
    leaves (or terminal) holding the "actual values".

    When wrapping a mapping in KeyPathMap, you can pretend that you have a flat mapping from (root to leaf) paths
    instead of a nested structure, and do your mapping CRUD with that view.

    >>> d = {'a': {
    ...         'a': '2a',
    ...         'b': {'a': 'aba',
    ...               'b': 3}
    ...         },
    ...      'c': 3.14
    ...     }
    >>> kp = KeyPathMap(d)
    >>> list(kp.items())
    [(('a', 'a'), '2a'), (('a', 'b', 'a'), 'aba'), (('a', 'b', 'b'), 3), (('c',), 3.14)]
    >>> list(kp)
    [('a', 'a'), ('a', 'b', 'a'), ('a', 'b', 'b'), ('c',)]
    >>> len(kp)
    4
    >>> assert list(kp) == list(kp.keys())
    >>> list(kp.values())
    ['2a', 'aba', 3, 3.14]
    >>> kp['a']
    {'a': '2a', 'b': {'a': 'aba', 'b': 3}}
    >>> kp[('a',)]
    {'a': '2a', 'b': {'a': 'aba', 'b': 3}}
    >>> kp['a', 'a']
    '2a'
    >>> kp['a', 'b', 'b']
    3
    >>> ('a', 'new_key') in kp
    False
    >>> kp['a', 'new_key'] = 'new val'
    >>> ('a', 'new_key') in kp
    True
    >>> kp['a', 'new_key']
    'new val'
    >>> len(kp)
    5
    >>> del kp['a', 'b', 'a']
    >>> len(kp)
    4
    >>> list(kp.items())
    [(('a', 'a'), '2a'), (('a', 'b', 'b'), 3), (('a', 'new_key'), 'new val'), (('c',), 3.14)]
    >>>
    >>> # By default, you can only write on already created nodes. But if auto_node_writes=True, you can do this:
    >>> kp = KeyPathMap(auto_node_writes=True)
    >>> kp
    {}
    >>> kp['a', 'b', 'c'] = 'hi world!'
    >>> kp
    {'a': {'b': {'c': 'hi world!'}}}
    """

    def __init__(
        self,
        store=dict,
        key_type: type = None,
        node_type: type = None,
        auto_node_writes=False,
    ):
        """
        Initialize a KeyPathMap.
        :param store: Your mapping, or the type of your mapping.
        :param key_type: The type of the keys
        :param node_type: The node type (typically the same as the store type
        :param auto_node_writes: False by default, which means you can only write on already created nodes.
            But if True, you can assign to paths with prefixes that don't yet exist. See examples.
        """
        if isinstance(store, type):
            store = store()
        store_type = type(store)
        if node_type is None:
            node_type = store_type
        if key_type is None:
            try:  # consider the type of the first key found in the store as the type for all
                key_type = type(next(iter(store.keys())))
            except StopIteration:
                key_type = str  # default to string keys

        def is_node(x):
            return isinstance(x, node_type)

        def is_key(x):
            return isinstance(x, key_type)

        self.store = store
        self.is_node = is_node
        self.is_key = is_key
        self.mk_new_node = node_type
        self.auto_node_writes = auto_node_writes

    def __getitem__(self, k):
        if not self.is_key(k):
            d = self.store.__getitem__(k[0])
            for kk in k[1:]:
                d = d[kk]
            return d
        else:
            return self.store.__getitem__(k)

    def __setitem__(self, k, v):
        if self.auto_node_writes:
            self._recursive_setitem(k, v)
        else:
            if not self.is_key(k) and len(k) > 1:
                d = self.store.__getitem__(k[0])
                for kk in k[1:-1]:
                    d = d[kk]
                d[k[-1]] = v
            else:
                self.store.__setitem__(k, v)

    def __delitem__(self, k):
        if not self.is_key(k):
            if len(k) == 1:
                self.store.__delitem__(k[0])
            else:
                d = self.store.__getitem__(k[0])
                for kk in k[1:-1]:
                    d = d[kk]
                d.__delitem__(k[-1])
        else:
            self.store.__delitem__(k)

    def items(self):
        return self._items(self.store)

    def __iter__(self):
        return map(lambda x: x[0], self.items())

    def __contains__(self, k):
        if self.is_key(k):
            return self.store.__contains__(k)
        else:
            d = self.store
            for kk in k:
                if not d.__contains__(kk):
                    return False
                else:
                    d = d[kk]
            return True  # if you got so far, you have all the keys in the path

    def __len__(self):
        c = 0
        for _ in self.items():
            c += 1
        return c

    def _recursive_setitem(self, k, v):
        if not self.is_key(k):
            d = self.store
            for kk in k[:-1]:
                if kk not in d:
                    d[kk] = self.mk_new_node()
                d = d[kk]
                # else:
                #     d = d[kk]
                #     assert self.is_node(d), f"Trying to set a value for path {k}, but {kk} was a leaf"
            d[k[-1]] = v
        else:
            self.store.__setitem__(k, v)

    @trans_generator_output(lambda x: (tuple(x[0]), x[1]))
    def _items(self, d, key_path_prefix=None):
        if key_path_prefix is None:
            for k, v in d.items():
                if not self.is_node(v):
                    yield [k], v
                else:
                    for kk, vv in self._items(v, [k]):
                        yield kk, vv
        else:
            for k, v in d.items():
                if not self.is_node(v):
                    yield key_path_prefix + [k], v
                else:
                    if self.is_key(k):
                        k = [k]
                    for kk, vv in self._items(v, k):
                        yield key_path_prefix + list(kk), vv

    def __repr__(self):
        return self.store.__repr__()

    def flat_dict(self):
        return {k: v for k, v in self.items()}

    def rollout(self, d):
        target = self.__class__(auto_node_writes=True)
        for k, v in d.items():
            target[k] = v
        return target.store

    def extract(self, key_paths=None, default_val=NO_DFLT, target=dict):
        if key_paths is None:
            key_paths = self.keys()
        if isinstance(target, type):
            target = target()
        for kp in key_paths:
            try:
                target[kp] = self[kp]
            except KeyError:
                if default_val is not NO_DFLT:
                    target[kp] = default_val
                else:
                    raise
        return target


class StrKeyPath(KeyPathMap):
    """
    A KeyPathMap, but where the key paths are expressed as string with a separator.
    If sep = '.', then instead of using ('a', 'b', 'c') as a key, you can use 'a.b.c'.

    >>> d = {'a': {
    ...         'a': '2a',
    ...         'b': {'a': 'aba',
    ...               'b': 3}
    ...         },
    ...      'c': 3.14
    ...     }
    >>> # Example with sep='/'
    >>> kp = StrKeyPath(d, sep='/')
    >>> list(kp.items())
    [('a/a', '2a'), ('a/b/a', 'aba'), ('a/b/b', 3), ('c', 3.14)]
    >>> # You can also add a prefix to the keys
    >>> kp = StrKeyPath(d, sep='/', prefix="http://")
    >>> list(kp.items())
    [('http://a/a', '2a'), ('http://a/b/a', 'aba'), ('http://a/b/b', 3), ('http://c', 3.14)]
    >>>
    >>> # Default sep is '.', so we'll work with that:
    >>> kp = StrKeyPath(d)
    >>> kp
    {'a': {'a': '2a', 'b': {'a': 'aba', 'b': 3}}, 'c': 3.14}
    >>> list(kp.items())
    [('a.a', '2a'), ('a.b.a', 'aba'), ('a.b.b', 3), ('c', 3.14)]
    >>> list(kp)
    ['a.a', 'a.b.a', 'a.b.b', 'c']
    >>> len(kp)
    4
    >>> assert list(kp) == list(kp.keys())
    >>> list(kp.values())
    ['2a', 'aba', 3, 3.14]
    >>> kp['a']
    {'a': '2a', 'b': {'a': 'aba', 'b': 3}}
    >>> kp['a.a']
    '2a'
    >>> kp['a.b.b']
    3
    >>> ('a.new_key') in kp
    False
    >>> kp['a.new_key'] = 'new val'
    >>> 'a.new_key' in kp
    True
    >>> kp['a.new_key']
    'new val'
    >>> len(kp)
    5
    >>> del kp['a.b.a']
    >>> len(kp)
    4
    >>> list(kp.items())
    [('a.a', '2a'), ('a.b.b', 3), ('a.new_key', 'new val'), ('c', 3.14)]
    >>>
    >>> # By default, you can only write on already created nodes. But if auto_node_writes=True, you can do this:
    >>> kp = StrKeyPath(auto_node_writes=True)
    >>> kp
    {}
    >>> kp['a.b.c'] = 'hi world!'
    >>> kp
    {'a': {'b': {'c': 'hi world!'}}}
    >>>
    """

    def __init__(
        self,
        store=dict,
        key_type: type = None,
        node_type: type = None,
        auto_node_writes=False,
        sep: str = '.',
        prefix: str = '',
    ):
        prefix_length = len(prefix)
        self.sep = sep
        self.prefix = prefix
        self._id_of_key = lambda k: tuple(k[prefix_length:].split(sep))
        self._key_of_id = lambda _id: prefix + sep.join(_id)
        super().__init__(
            store=store,
            key_type=key_type,
            node_type=node_type,
            auto_node_writes=auto_node_writes,
        )

    def __getitem__(self, k):
        return super().__getitem__(self._id_of_key(k))

    def __setitem__(self, k, v):
        return super().__setitem__(self._id_of_key(k), v)

    def __delitem__(self, k):
        return super().__delitem__(self._id_of_key(k))

    def __contains__(self, k):
        return super().__contains__(self._id_of_key(k))

    def items(self):
        yield from ((self._key_of_id(_id), v) for _id, v in super().items())

    def rollout(self, d):
        target = self.__class__(sep=self.sep, prefix=self.prefix, auto_node_writes=True)
        for k, v in d.items():
            target[k] = v
        return target.store


# TODO: Not what is below should remain. Probably better to just use the above.
class KeyPathTrans:
    """
    Doing what StrKeyPath but where the store that is being operated on is not included in the object, but
    given to the method as input.
    """

    def __init__(self, sep: str = '.', node_type: type = dict, mk_new_node=None):
        """

        :param sep:
        :param node_type:
        """
        self.sep = sep
        self.node_type = node_type
        if mk_new_node is None:
            mk_new_node = node_type
        self.mk_new_node = mk_new_node

    def items(self, d, key_path_prefix=None):
        """
        iterate through items of store recursively, yielding (key_path, val) pairs for all nested values that are not
        store types.
        That is, if a value is a store_type, it won't generate a yield, but rather, will be iterated through
        recursively.
        :param d: input store
        :param key_path_so_far: string to be prepended to all key paths (for use in recursion, not meant for direct use)
        :return: a (key_path, val) iterator

        >>> kp = KeyPathTrans()
        >>> input_dict = {
        ...     'a': {
        ...         'a': 'a.a',
        ...         'b': 'a.b',
        ...         'c': {
        ...             'a': 'a.c.a'
        ...         }
        ...     },
        ...     'b': 'b',
        ...     'c': 3
        ... }
        >>> list(kp.items(input_dict))
        [('a.a', 'a.a'), ('a.b', 'a.b'), ('a.c.a', 'a.c.a'), ('b', 'b'), ('c', 3)]
        """
        if key_path_prefix is None:
            for k, v in d.items():
                if not isinstance(v, self.node_type):
                    yield k, v
                else:
                    for kk, vv in self.items(v, k):
                        yield kk, vv
        else:
            for k, v in d.items():
                if not isinstance(v, self.node_type):
                    yield key_path_prefix + self.sep + k, v
                else:
                    for kk, vv in self.items(v, k):
                        yield key_path_prefix + self.sep + kk, vv

    def keys(self, d):
        for k, v in d.items():
            if not isinstance(v, self.node_type):
                yield k
                # key_path_list.append(k)
            else:
                yield from (k + self.sep + x for x in self.keys(v))

    def getitem(self, d, key_path, default_val=None):
        """
        getting with a key list or "."-separated string
        :param d: dict-like
        :param key_path: list or "."-separated string of keys
        :return:
        """
        if isinstance(key_path, str):
            key_path = key_path.split(self.sep)
        try:
            return reduce(operator.getitem, key_path, d)
        except (TypeError, KeyError):
            return default_val

    def setitem(self, d, key_path, val):
        """
        setting with a key list or "."-separated string
        :param d: dict
        :param key_path: list or "."-separated string of keys
        :param val: value to assign
        :return:
        """
        if isinstance(key_path, str):
            key_path = key_path.split(self.sep)
        self.getitem(d, key_path[:-1])[key_path[-1]] = val

    def setitem_recursive(self, d, key_path, val):
        """

        :param d:
        :param key_path:
        :param val:
        :return:

        >>> kp = KeyPathTrans()
        >>> input_dict = {
        ...   "a": {
        ...     "c": "val of a.c",
        ...     "b": 1,
        ...   },
        ...   "10": 10,
        ...   "b": {
        ...     "B": {
        ...       "AA": 3
        ...     }
        ...   }
        ... }
        >>>
        >>> kp.setitem_recursive(input_dict, 'new.key.path', 7)
        >>> input_dict
        {'a': {'c': 'val of a.c', 'b': 1}, '10': 10, 'b': {'B': {'AA': 3}}, 'new': {'key': {'path': 7}}}
        >>> kp.setitem_recursive(input_dict, 'new.key.old.path', 8)
        >>> input_dict
        {'a': {'c': 'val of a.c', 'b': 1}, '10': 10, 'b': {'B': {'AA': 3}}, 'new': {'key': {'path': 7, 'old': {'path': 8}}}}
        >>> kp.setitem_recursive(input_dict, 'new.key', 'new val')
        >>> input_dict
        {'a': {'c': 'val of a.c', 'b': 1}, '10': 10, 'b': {'B': {'AA': 3}}, 'new': {'key': 'new val'}}
        """
        if isinstance(key_path, str):
            key_path = key_path.split(self.sep)
        first_key = key_path[0]
        if len(key_path) == 1:
            d[first_key] = val
        else:
            if first_key in d:
                self.setitem_recursive(d[first_key], key_path[1:], val)
            else:
                d[first_key] = self.node_type()
                self.setitem_recursive(d[first_key], key_path[1:], val)

    def extract_key_paths(
        self, d, key_paths, field_naming='full', use_default=False, default_val=None,
    ):
        """
        getting with a key list or "."-separated string
        :param d: dict-like
        :param key_path: list or "."-separated string of keys
        :param field_naming: 'full' (default) will use key_path strings as is, leaf will only use the last dot item
            (i.e. this.is.a.key.path will result in "path" being used)
        :return:

        >>> kp = KeyPathTrans()
        >>> d = {
        ...     'a': {
        ...         'a': 'a.a',
        ...         'b': 'a.b',
        ...         'c': {
        ...             'a': 'a.c.a'
        ...         }
        ...     },
        ...     'b': 'b',
        ...     'c': 3
        ... }
        >>> kp.extract_key_paths(d, 'a')
        {'a': {'a': 'a.a', 'b': 'a.b', 'c': {'a': 'a.c.a'}}}
        >>> kp.extract_key_paths(d, 'a.a')
        {'a.a': 'a.a'}
        >>> kp.extract_key_paths(d, 'a.c')
        {'a.c': {'a': 'a.c.a'}}
        >>> kp.extract_key_paths(d, ['a.a', 'a.c'])
        {'a.a': 'a.a', 'a.c': {'a': 'a.c.a'}}
        >>> kp.extract_key_paths(d, ['a.a', 'something.thats.not.there'])  # missing key just won't be included
        {'a.a': 'a.a'}
        >>> kp.extract_key_paths(d, ['a.a', 'something.thats.not.there'], use_default=True, default_val=42)
        {'a.a': 'a.a', 'something.thats.not.there': 42}
        """
        dd = self.mk_new_node()
        if isinstance(key_paths, str):
            key_paths = [key_paths]
        if isinstance(key_paths, self.node_type):
            key_paths = [k for k, v in key_paths.items() if v]

        for key_path in key_paths:

            if isinstance(key_path, str):
                field = key_path
                key_path = key_path.split(self.sep)
            else:
                field = self.sep.join(key_path)

            if field_naming == 'leaf':
                field = key_path[-1]
            else:
                field = field

            try:
                dd.update({field: reduce(operator.getitem, key_path, d)})
            except (TypeError, KeyError):
                if use_default:
                    dd.update({field: default_val})

        return dd


# from collections.abc import MutableMapping
#
#
# class AttrStore(MutableMapping):
#     def __getitem__(self, k):
#         if not hasattr(self, k):
#             setattr(self, k, AttrStore())
#         return getattr(self, k)
#
#     def __setitem__(self, k, v):
#         return setattr(self, k, v)
#
#     def __delitem__(self, k):
#         return delattr(self, k)
#
#     def __iter__(self):
#         return iter(self.__dict__)
#
#     def keys(self):
#         return self.__dict__.keys()
#
#     def __contains__(self, k):
#         return hasattr(self, k)
#
#     def __len__(self):
#         return len(list(self))
#
#
# d = AttrStore()
# kp = KeyPathMap(d)
# kp['a'] = '2a'
# print(kp.store.a)
# print(list(kp))
# kp['c']['d'] = 3
# print(list(kp))
# kp['c', 'z'] = 'hi world'
# print(list(kp.items()))

## Not working yet:
# d = AttrStore()
# kp = StrKeyPath(d, sep='.')
# kp['a'] = '2a'
# print(kp.store.a)
# print(list(kp))
# kp['c']['d'] = 3
# print(list(kp))
# kp['c.z'] = 'hi world'
# list(kp.items())



from i2 import Sig
from i2.signatures import ch_variadics_to_non_variadic_kind


def foo(a, *args, bar, **kwargs):
    return f'{a=}, {args=}, {bar=}, {kwargs=}'


assert str(Sig(foo)) == '(a, *args, bar, **kwargs)'
wfoo = ch_variadics_to_non_variadic_kind(foo)
str(Sig(wfoo))
#'(a, args=(), *, bar, kwargs={})'

# And now to do this:

foo(1, 2, 3, bar=4, hello='world')
# "a=1, args=(2, 3), bar=4, kwargs={'hello': 'world'}"

# We can do it like this instead:

wfoo(1, (2, 3), bar=4, kwargs=dict(hello='world'))
# "a=1, args=(2, 3), bar=4, kwargs={'hello': 'world'}"

# Note, the outputs are the same. It's just the way we call our function that has
# changed.

assert wfoo(1, (2, 3), bar=4, kwargs=dict(hello='world')) == foo(
    1, 2, 3, bar=4, hello='world'
)
assert wfoo(1, (2, 3), bar=4) == foo(1, 2, 3, bar=4)
assert wfoo(1, (), bar=4) == foo(1, bar=4)

# Note that if there is not variadic positional arguments, the variadic keyword
# will still be a keyword-only kind.


@ch_variadics_to_non_variadic_kind
def func(a, bar=None, **kwargs):
    return f'{a=}, {bar=}, {kwargs=}'


str(Sig(func))
#'(a, bar=None, *, kwargs={})'
assert (
    func(1, bar=4, kwargs=dict(hello='world'))
    == "a=1, bar=4, kwargs={'hello': 'world'}"
)

# If the function has neither variadic kinds, it will remain untouched.


def func(a, /, b, *, c=3):
    return a + b + c


ch_variadics_to_non_variadic_kind(func) == func
# True

# If you only want the variadic positional to be handled, but leave leave any
# VARIADIC_KEYWORD kinds (**kwargs) alone, you can do so by setting
# `ch_variadic_keyword_to_keyword=False`.
# If you'll need to use `ch_variadics_to_non_variadic_kind` in such a way
# repeatedly, we suggest you use `functools.partial` to not have to specify this
# configuration repeatedly.

from functools import partial

tuple_the_args = partial(
    ch_variadics_to_non_variadic_kind, ch_variadic_keyword_to_keyword=False
)


@tuple_the_args
def foo(a, *args, bar=None, **kwargs):
    return f'{a=}, {args=}, {bar=}, {kwargs=}'


Sig(foo)
# <Sig (a, args=(), *, bar=None, **kwargs)>
foo(1, (2, 3), bar=4, hello='world')
# "a=1, args=(2, 3), bar=4, kwargs={'hello': 'world'}"
